@echo off

echo "Force S/UNI-622 out of reset, in case someone screwed it up"
memhelp 0xfebff000 8 0x00

rem echo "Read from S/UNI-622 address before write (kluge for bad address latching)"
rem memhelp 0xfebff140 32

echo "Set up S/UNI 622 Cell Receiver to pass all (including idle) cells"
memhelp 0xfebff140 32 0xa4

echo "Set-up AMCC Bus master"
rem assuming AMCC range starts at phys 0xfebfff40

rem AMCC bus master control/status register is at offset 0x3C
rem value 0x04000000 means:
rem     bits 31:29  0 = disable nvRAM access from PCI
rem     bit 28      0 = reserved
rem     bit 27      0 = do not reset mailbox status flags
rem     bit 26      1 = reset addon-to-PCI FIFO status flags
rem     bit 25      0 = do not reset PCI-to-addon FIFO status flags
rem     bit 24      0 = do not assert addon reset pin
rem     bits 23:16  0 = nvRAM address/data port
rem     bit 15      0 = reserved
rem     bit 14      0 = disable read transfers
rem     bit 13      0 = read FIFO .GT. 0 words cause PCI bus request
rem     bits 12,8  00 = write transfers have priority over reads
rem     bit 11      0 = reserved
rem     bit 10      0 = disable write transfers
rem     bit 9       0 = write FIFO .GT. 0 words cause PCI bus request
rem     bit 7       0 = (read only) write transfer count is not all zeroes
rem     bit 6       0 = (read only) read  transfer count is not all zeroes
rem     bit 5       0 = addon-to-PCI FIFO not empty
rem     bit 4       0 = addon-to-PCI FIFO not 4+ filled
rem     bit 3       0 = addon-to-PCI FIFO not full
rem     bit 2       0 = PCI-to-addon FIFO not empty
rem     bit 1       0 = PCI-to-addon FIFO not 4+ spaces
rem     bit 0       0 = PCI-to-addon FIFO not full
memhelp 0xfebfff7c 32 0x04000000

rem AMCC PCI-controlled bus master write address register is at offset 0x24
rem next card write will occur above 7 MB + 64 KB line
rem
rem because DOS + High Memory Area (HMA) leaves off at 1 MB + 64 KB line
rem and SMARTDRV uses another 2 MB
rem and Borland C++ DPMI32 server uses another 4 MB
rem     when environment variable DPMIMEM has value "MAXMEM 4096"
memhelp 0xfebfff64 32 0x00710000

rem AMCC PCI-controlled bus master write transfer count register is at offset 0x28
rem next card write will be allowed to use 48 MB
memhelp 0xfebfff68 32 0x03000000

rem AMCC bus master control/status register is at offset 0x3C
rem value 0x00000400 means same as 0x04000000 above, except:
rem     bit 26      0 = do not reset addon-to-PCI FIFO status flags
rem     bit 10      1 = enable write transfers
memhelp 0xfebfff7c 32 0x00000400

rem assuming RX FPGA range starts at 0xFEBFF400

echo "clear latched RX errors"
rem     bits 7:2    1 = reserved
rem     bit 1       1 = (readonly cleared) AIM622 data ever had parity error
rem     bit 0       1 = (readonly cleared) external FIFO was ever full
memhelp 0xfebff404 32

echo "Release external FIFOs reset"
rem     bit 7       0 = (readonly) external FIFO not full
rem     bit 6       0 = (readonly) external FIFO not almost (.LE. 8 words remaining) full
rem     bit 5       0 = (readonly) external FIFO not almost (.LE. 8 words) empty
rem     bit 4       0 = (readonly) external FIFO not empty
rem     bit 3       0 = disable AMCC writes
rem     bit 2       0 = disable external FIFO writes
rem     bit 1       0 = disable reads from S/UNI 622
rem     bit 0       1 = disable reset to external FIFO
memhelp 0xfebff400 32 0x01

echo "Enable Utopia Cell Reception"
rem same as above, except:
rem     bit 1       1 = enable reads from S/UNI 622
memhelp 0xfebff400 32 0x03

echo "Enable writes to AMCC FIFO"
rem same as above, except:
rem     bit 3       1 = enable AMCC writes
memhelp 0xfebff400 32 0x0b

echo "Enable writes to external FIFOs"
rem same as above, except:
rem     bit 2       0 = enable external FIFO writes
memhelp 0xfebff400 32 0x0f
