#include <dos.h>
#include <stdlib.h>                 // malloc()
#include <mem.h>                    // memcpy()
#include "..\types\types.h"
#include "..\dpmi32\intr.h"



#ifndef __DPMI32__
#    error Must be compiled for 32-bit DPMI.  Use BCC32 -WX
#endif

#pragma inline



// DPMI call to set protected mode interrupt vector
//
void setvect( Bit8 vector, Bit48 pointer )
{
    union REGS regs;

    regs.w.flags = 0;
    regs.w.ax = 0x205;
    regs.h.bl = vector;
    regs.w.cx = pointer.selector;
    regs.x.edx = pointer.offset;
    int386( 0x31, &regs, &regs );
}



// DPMI call to set real mode interrupt vector
//
void setvect_real( Bit8 vector, Bit32 pointer )
{
    union REGS regs;

    regs.w.flags = 0;
    regs.w.ax = 0x201;
    regs.h.bl = vector;
    regs.w.cx = pointer >> 16;
    regs.w.dx = pointer & 65535;
    int386( 0x31, &regs, &regs );
}



// DPMI call to get protected mode interrupt vector
//
Bit48 getvect( Bit8 vector )
{
    union REGS regs;
    Bit48 pointer;

    regs.w.flags = 0;
    regs.w.ax = 0x204;
    regs.h.bl = vector;
    int386( 0x31, &regs, &regs );

    pointer.offset = regs.x.edx;
    pointer.selector = regs.w.cx;
    return pointer;
}



// DPMI call to get real mode interrupt vector
//
Bit32 getvect_real( Bit8 vector )
{
    union REGS regs;

    regs.w.flags = 0;
    regs.w.ax = 0x200;
    regs.h.bl = vector;
    int386( 0x31, &regs, &regs );

    return (regs.w.cx << 16) | regs.w.dx;
}



extern  char    interrupt32Thunk[];
extern  int _interrupt32ThunkSize;
extern  int _I32TparmLoadLabel;
extern  int _I32TdsLoadLabel;
extern  int _I32TcallLabel;
extern  int _I32TchainLabel;

// all symbols defined in the assembler look like
// addresses of variables, so we must use the C
// address-of operator to use the offsets...
//
// otherwise, we would end up referencing memory
// at the locations like "54" and "76", which
// would certainly cause a protection violation!
//
#define interrupt32ThunkSize    ((int)&_interrupt32ThunkSize)
#define I32TparmLoadLabel       ((int)&_I32TparmLoadLabel)
#define I32TdsLoadLabel         ((int)&_I32TdsLoadLabel)
#define I32TcallLabel           ((int)&_I32TcallLabel)
#define I32TchainLabel          ((int)&_I32TchainLabel)

#define stampDD(where,what) *(void**)(thunk + (where)) = (what)
#define stampDW(where,what) *(unsigned short *)(thunk + (where)) = (what)



// allocates a small block of memory and copies code
// from a template into it, then patches it with the
// current value of the DS register, the address of your
// function, a pointer to pass to your function,
// and the address of your variable which holds the 48-bit
// address of the old interupt handler
//
// returns the address of the thunking code if successful
// or NULL if failure (could not allocate menory)
//
// assumptions:
//     DS and CS selectors alias the same physical memory
//     and will continue to do so at interrupt time
//
// your functions should accept arguments for all registers
// just like a C function in the 16-bit compilers for the 386
// processor used to
//
// if your function returns a nonzero value, the thunk will
// chain (not call -- stack has no extra stuff on it) the
// old interrupt handler
//
vfp make_thunk( ifp proc, Bit48 *chainAddr, void *parms )
{
    char *thunk;

    thunk = (char *) malloc( interrupt32ThunkSize );
    if( thunk )
    {
        memcpy( thunk, interrupt32Thunk, interrupt32ThunkSize );
        stampDD( I32TparmLoadLabel, parms );
        stampDW( I32TdsLoadLabel, _DS );
        stampDD( I32TcallLabel, (char *)(((char *)proc) - (thunk + I32TcallLabel + 4)) );
        stampDD( I32TchainLabel, chainAddr );
    }

    return (vfp) thunk;
}
