#include <stdio.h>              // printf()
#include <stdlib.h>             // exit()
#include "phys.h"



int main( int argc, char *argv[] )
{
    Bit32 handle;
    Bit32 address;
    Bit32 old_value, new_value;
    void *pointer;
    int width;
    int want_write;
    const char *program_name = argv[0];

    // parse command-line arguments
    //
    if( argc<3 || argc>4 )
    {
        fprintf( stderr, "usage: %s phys-address bit-width {new-value}\n", program_name );
        return 1;
    }
    address = strtoul( argv[1], NULL, 0 );
    width = atoi( argv[2] );

    if( argc == 4 )
    {
        new_value = strtoul( argv[3], NULL, 0 );
        want_write = TRUE;
    }
    else
        want_write = FALSE;

    // map one page of memory starting at the
    // physical address that the user specified
    //
    pointer = map_physical( address, 4, &handle );
    if( pointer == NULL )
    {
        printf( "Sorry, cannot map 4 bytes at physical address 0x%.8X to a flat pointer\n", PAGE_SIZE, address );
        return 2;
    }

    switch( width )
    {
        default:
            fprintf( stderr, "Sorry, you can only use widths of 8, 16, or 32 bits.\n" );
            unmap_physical( PAGE_SIZE, handle );
            return 3;

        case 8:
            if( want_write )
                * (Bit8 *) pointer = new_value;
            else
                old_value = * (Bit8 *) pointer;
            break;

        case 16:
            if( want_write )
                * (Bit16 *) pointer = new_value;
            else
                old_value = * (Bit16 *) pointer;
            break;

        case 32:
            if( want_write )
                * (Bit32 *) pointer = new_value;
            else
                old_value = * (Bit32 *) pointer;
            break;
    }

    // unmap the physical memory we mapped above
    //
    unmap_physical( PAGE_SIZE, handle );

    printf( "dealing with memory space %d bits at a time\n", width );

    if( want_write )
        printf( "wrote new value 0x%.*X at address 0x%.4X\n", width/4, new_value, address );
    else
        printf( "read old value 0x%.*X at address 0x%.4X\n", width/4, old_value, address );

    return 0;
}
