/* results:
 *
 * at least from 32-bit code, disk I/O is 40 times slower
 * if you do not write an even multiple of 512 bytes
 * (the sector size)
 *
 * YUCK!
 *
 *
 *
 * DOS writes to my ST32140A disk can only support 3.3 MB/S
 * right next door to the cylinders (440-449) that BIOS
 * (see disktst2.cpp) can support 4.4 MB/S
 *
 * I wonder if avoiding BIOS would be even faster?
 */




#include <errno.h>                  // extern errno _sys_errlist
#include <sys/stat.h>               // access modes for open() and creat()
#include <fcntl.h>                  // extern _fmode
#include <io.h>                     // creat() chsize() write()
#include <dos.h>                    // abswrite() getverify() setverify()
#include <stdio.h>                  // printf()
#include <stdlib.h>                 // exit()
#include <memory.h>                 // memset()
#include "../types/types.h"         // typedef Bit8 Bit32
#include "../intel/timestmp.h"      // class PentiumClock
#include "../dpmi32/flat.h"         // align()
#include "../dpmi32/realmode.h"     // real_to_physical() physical_to_real()







#define SIZE_WANTED     (10L*1024*1024)

#define BUFFER_ALIGNMENT    65536

//#define BLOCK_SIZE      (BUFFER_SIZE * COMPL_RING_ENTRIES)
//#define BLOCK_SIZE      (12 * 87381)
//#define BLOCK_SIZE      (1048572)     // BAD
//#define BLOCK_SIZE      (1228800)     // OK
//#define BLOCK_SIZE      (1048577)     // BAD
//#define BLOCK_SIZE      (1200000)     // BAD
//#define BLOCK_SIZE      (1230000)     // BAD
//#define BLOCK_SIZE      (1228812)     // BAD
//#define BLOCK_SIZE      (1229312)     // OK
//#define BLOCK_SIZE      (1044992)
//#define BLOCK_SIZE      (1044480)
//#define BLOCK_SIZE      (64*1024)
//#define BLOCK_SIZE      (32*1024)
#define BLOCK_SIZE      (16*1024)

#define SECTOR_SIZE     512
#define SECTOR_COUNT    (BLOCK_SIZE / SECTOR_SIZE)

#define ITERATIONS      1000
//#define ITERATIONS      121


#if __FLAT__
#define huge
#endif
Bit8 huge buffer0[ BLOCK_SIZE + BUFFER_ALIGNMENT - 1 ];
Bit8 huge *buffer;



int main( int argc, char *argv[] )
{
    const char *program = argv[ 0 ];
    const char *filename = argv[ 1 ];
    PentiumClock start, end;
    double timing;
    int result, handle;
    unsigned count;
    long dos_sector;
    int head, track, sector;
    int head_count, track_count, sector_count;

    if( argc != 2 )
    {
        fprintf( stderr, "usage: %s filename\n", program );
        return 1;
    }

#if 0
    // seek to the 10 MB position in the file
    //
    // not implemented yet!
    //

    // set the length of the file to 10 MB
    //
    start.set();
    result = chsize( handle, SIZE_WANTED );
    end.set();
    if( result != 0 )
    {
        printf( "error %d '%s' setting file size to %ld\n", errno, _sys_errlist[errno], SIZE_WANTED );
        exit( 2 );
    }
    printf( "set size to %ld\n", SIZE_WANTED );
    timing = end - start;
    printf( "elapsed time = %g Pentium clocks\n", timing );
    printf( "\n" );
#endif






    // cause the buffer pointer to be aligned
    //
#if __FLAT__
    buffer = (Bit8 *) align( BUFFER_ALIGNMENT, buffer0 );
#else
    buffer = (Bit8 huge *) physical_to_real( (Bit32)align( BUFFER_ALIGNMENT, (void *)real_to_physical( (Bit32)buffer0 ) ) );
#endif
    printf( "raw buffer is at 0x%p\n", buffer0 );
    printf( "aligned %lu byte buffer is at 0x%p\n", (Bit32)BLOCK_SIZE, buffer );

    // fill the buffer with binary zero
    //
    memset( buffer, 0, BLOCK_SIZE );







    // this section uses DOS interrupt 0x21
    //

    // create a file on the 2nd hard disk
    //
    _fmode = O_BINARY;
    start.set();
    handle = creat( filename, S_IWRITE );
    end.set();
    if( handle < 0 )
    {
        printf( "error %d '%s' creating file\n", errno, _sys_errlist[errno] );
        exit( 1 );
    }
    printf( "created file; handle is %d\n", handle );
    timing = end - start;
    timing /= CLOCK_RATE;
    printf( "elapsed time = %g seconds\n", timing );

    // tell the user how the DOS verify flag is set,
    // and turn off verification for this program
    //
    // won't run in Windoze 95
    //
//    printf( "DOS verify (readback of sectors) is currently %s\n", getverify()?"on":"off" );

    printf( "writing %d bytes of binary zero to '%s' %d times starting at logical file offset\n", BLOCK_SIZE, filename, ITERATIONS );

    // write the buffer to consecutive file bytes
    // lots of times
    //
    start.set();

    for( count=0; count<ITERATIONS; count++ )
    {
        result = write( handle, buffer, BLOCK_SIZE );
        if( result < 0 )
        {
            printf( "C library error %d '%s' doing write() #%d to '%s' for length %d bytes\n", errno, _sys_errlist[errno], count, filename, BLOCK_SIZE );
            exit( 2 );
        }
    }

    end.set();
    printf( "done\n" );

    end -= start;
    timing = end.get_seconds();

    printf( "total elapsed time = %g seconds\n", timing );
    printf( "average time = %g seconds per %lu byte buffer\n", timing/ITERATIONS, (Bit32)BLOCK_SIZE );
    printf( "average %g bytes per second\n", (Bit32)BLOCK_SIZE*ITERATIONS/timing );

    printf( "closing file\n" );
    start.set();
    close( handle );
    end.set();
    end -= start;
    timing = end.get_seconds();
    printf( "elapsed time = %g seconds\n", timing );

    printf( "\n" );

    return 0;
}
