/* conditional compilation switches
 */
#define FORCE_LARGE_FILES       FALSE
#define ALLOW_DISK_WRITE        TRUE
#define DEBUG_VARIANCE          FALSE
#define REPORT_OVERHEAD_STATS   FALSE
#define DEBUG_EXTRA_TCP_STATS   FALSE



/* manifest constants
 */
#define BIG_BYTES_BREAKPOINT    (1024*1024L)



/* dostats.c */
/* compute global stats on flows */
/* jmm & kcc 29 Jul 95 */

#define MAX_PROTOCOL 255
#define MAX_PORT 65536

#if __unix__
#   include <sys/types.h>
#   include <sys/time.h>
#   include <netinet/in.h>
#else
#   include "..\wattcp\tcp.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#if __unix__
#   include "types.h"
#else
#   include "..\types\types.h"
#   include "..\intel\byteswap.h"
#endif

#include "queue.h"
#include "trace.h"              /* must be included before util.h */
#include "util.h"
#include "dostats.h"
#include "fragstat.h"
#include "portstat.h"
#include "netstats.h"
#include "hoststat.h"



// none of this code gets compiled in
// if it's isn't going to be called
//
#if DO_FOREGROUND_STATS



static Bit32    flows;
static double   packetsum, bytesum, dursum;
static double   packet2sum, byte2sum, dur2sum;
static double   sumpb, sumpd, sumbd;
static int      proto_packets[MAX_PROTOCOL+1],
                proto_flows[MAX_PROTOCOL+1];
static double   proto_bytes[MAX_PROTOCOL+1];
static double   proto_duration[MAX_PROTOCOL+1];
/* arrays for variance/covar */
static double   sum2_packets[MAX_PROTOCOL+1], sum2_bytes[MAX_PROTOCOL+1],
                sum2_duration[MAX_PROTOCOL+1];
static double   sum_pb[MAX_PROTOCOL+1], sum_bd[MAX_PROTOCOL+1],
                sum_pd[MAX_PROTOCOL+1];
#if EXTRA_TCP_STATS
/* largest value the logarithm (base 2)
 * of a 32-bit unsigned integer can be:
 *
 * 0 for no bits set
 * 1 to 32 for 1 to 32 bits set
 */
#define BIT32_LOG2_MAX      32

#if CELLS_PER_PACKET > 1
/* how many bits I have devoted to my bitmapped
 * representation of the freeform TCP options,
 * and how many permutations those bits can have
 */
#define TCP_OPTION_BITS     12
#define TCP_OPTION_MAX      (1 << TCP_OPTION_BITS)
#endif

static Bit32    tcp_flows;
static double   tcp_bytes;
static double   tcp_packets;
static double   tcp_ack_bytes;
static double   tcp_seq_bytes;
static Bit32    tcp_seq_small_flows_leftover;
static double   tcp_seq_small_bytes_leftover;
static double   tcp_seq_small_packets_leftover;
static Bit32    tcp_seq_big_flows_leftover;
static double   tcp_seq_big_bytes_leftover;
static double   tcp_seq_big_packets_leftover;

/* these are indexed by the field in flowstats
 * with the name "fs_" + name of array
 *
 * these are counts of flows
 */
static Bit32      tcp_seq_n_holes[ BIT8_MAX + 1 ];
static Bit32      tcp_seq_1_holes[ BIT8_MAX + 1 ];
static Bit32    tcp_seq_hole_pkts[ BIT8_MAX + 1 ];
static Bit32        tcp_seq_backs[ BIT8_MAX + 1 ];
static Bit32         tcp_seq_dups[ BIT8_MAX + 1 ];
static Bit32         tcp_ack_dups[ BIT8_MAX + 1 ];
static Bit32       tcp_bad_length[ BIT8_MAX + 1 ];
static Bit32            tcp_flags[ BIT8_MAX + 1 ];
static Bit32          tcp_win_max[ BIT32_LOG2_MAX + 1 ];     /* uses log base 2 */
#if CELLS_PER_PACKET > 1
static Bit32          tcp_options[ TCP_OPTION_MAX ];
static Bit32      tcp_max_segment[ BIT16_MAX + 1 ];
static Bit32        tcp_win_scale[ TCP_MAX_WINSHIFT + 2 ];
#endif

/* this holds strings that represent all possible
 * combinations of the 8 bits in TCP flags field
 */
static char tcp_flags_string[ BIT8_MAX + 1 ][ 9 ];
static int is_tcp_flags_string_filled = FALSE;

#if CELLS_PER_PACKET > 1
/* this holds strings that represent all possible
 * combinations of the 8 bits in (my) TCP options field
 */
static char tcp_options_string[ TCP_OPTION_MAX ][ TCP_OPTION_BITS + 1 ];
static int is_tcp_options_string_filled = FALSE;
#endif  // CELLS_PER_PACKET>1
#endif  // EXTRA_TCP_STATS




// aha!  if you don't convert n to floating point
// before squaring it, its 32 bits will overflow
// if it was greater than 64K
//
double variance(int n, double sum, double sum2)
{
#if !ALLOW_INFINITE_RESULTS
    if( n == 0 )
        return 0;
#endif

    return ((n * sum2) - (sum * sum))
    /      ( (double)n * (n>1 ? n-1 : 1) );
}



// forgets all statistics
//
int init_stats( void )
{
    Bit32 count;
    int result;

    flows = 0;
    packetsum = 0;
    bytesum = 0;
    packet2sum = 0;
    byte2sum = 0;
    dursum = 0;
    dur2sum = 0;
    sumpb = 0;
    sumpd = 0;
    sumbd = 0;

#if EXTRA_TCP_STATS
    tcp_flows = 0;
    tcp_bytes = 0;
    tcp_packets = 0;
    tcp_ack_bytes = 0;
    tcp_seq_bytes = 0;
    tcp_seq_big_flows_leftover = 0;
    tcp_seq_big_bytes_leftover = 0;
    tcp_seq_big_packets_leftover = 0;
    tcp_seq_small_flows_leftover = 0;
    tcp_seq_small_bytes_leftover = 0;
    tcp_seq_small_packets_leftover = 0;

    for( count=0; count<=BIT8_MAX; count++ )
    {
        tcp_seq_hole_pkts[ count ] = 0;
        tcp_seq_n_holes[ count ] = 0;
        tcp_seq_1_holes[ count ] = 0;
        tcp_bad_length[ count ] = 0;
        tcp_seq_backs[ count ] = 0;
        tcp_seq_dups[ count ] = 0;
        tcp_ack_dups[ count ] = 0;
        tcp_flags[ count ] = 0;
    }

    for( count=0; count<=BIT32_LOG2_MAX; count++ )
    {
        tcp_win_max[ count ] = 0;
    }

#if CELLS_PER_PACKET > 1
    for( count=0; count<=BIT16_MAX; count++ )
    {
        tcp_max_segment[ count ] = 0;
    }

    for( count=0; count<TCP_MAX_WINSHIFT+2; count++ )
    {
        tcp_win_scale[ count ] = 0;
    }

    for( count=0; count<TCP_OPTION_MAX; count++ )
    {
        tcp_options[ count ] = 0;
    }

    if( !is_tcp_options_string_filled )
    {
        is_tcp_options_string_filled = TRUE;

        for( count=0; count<TCP_OPTION_MAX; count++ )
        {
            sprintf( tcp_options_string[count], "%s%s%s%s%s%s%s%s%s%s%s%s",
                count & fs_tcp_opt_reserved       ? "R" : "",
                count & fs_tcp_opt_ttcp_cc_echo   ? "E" : "",
                count & fs_tcp_opt_ttcp_cc_new    ? "N" : "",
                count & fs_tcp_opt_ttcp_cc        ? "C" : "",
                count & fs_tcp_opt_too_short      ? "H" : "",
                count & fs_tcp_opt_truncated      ? "D" : "",
                count & fs_tcp_opt_unknown        ? "U" : "",
                count & fs_tcp_opt_timestamp      ? "T" : "",
                count & fs_tcp_opt_max_segment    ? "M" : "",
                count & fs_tcp_opt_win_scale      ? "W" : "",
                count & fs_tcp_opt_sack           ? "S" : "",
                count & fs_tcp_opt_sack_permitted ? "P" : ""
            );
        }
    }
#endif

    if( !is_tcp_flags_string_filled )
    {
        is_tcp_flags_string_filled = TRUE;

        for( count=0; count<256; count++ )
        {
            sprintf( tcp_flags_string[count], "%s%s%s%s%s%s%s%s",
                count & 128     ? "8" : "",
                count & 64      ? "7" : "",
                count & TH_URG  ? "U" : "",
                count & TH_ACK  ? "A" : "",
                count & TH_PUSH ? "P" : "",
                count & TH_RST  ? "R" : "",
                count & TH_SYN  ? "S" : "",
                count & TH_FIN  ? "F" : ""
            );
        }
    }
#endif

    for( count=0; count<=MAX_PROTOCOL; count++ )
    {
        proto_flows[count]=0;
        proto_bytes[count]=0;
        proto_packets[count]=0;
        proto_duration[count]=0;
        sum2_duration[count]=0;
        sum2_packets[count]=0;
        sum2_bytes[count]=0;
        sum_pb[count]=0;
        sum_bd[count]=0;
        sum_pd[count]=0;
    }

#if REASM_IP_FRAGS
    result = init_fragment_stats();
    if( result )
        return result;
#endif

    result = init_port_stats();
    if( result )
        return result;

    result = init_host_stats();
    if( result )
        return result;

    return init_net_stats();
}



/* Compute the following statistics:
 * mean and variance of packets, bytes, and duration
 * covariance
 *
 * believe it or not, calculating the number of TCP
 * bytes in either seqnum or acknum by subtracting
 * min from max using 32-bit unsigned math, and ignoring
 * carry, actally gives us the correct value, both when
 * SEQ/ACK has wrapped from 4G to zero and when it hasn't!
 */
void update_stats( struct hashrec *hr, float duration )
{
    u_char proto = hr->hr_fd.fd_proto;
    u_int  packets = hr->hr_packets;
    Bit64  bytes = hr->hr_bytes;
    double packets2 = (double)packets * packets;
    double bytes2 = (double)bytes * bytes;
    double duration2 = (double)duration * duration;
    double packets_bytes = (double)packets * bytes;
    double bytes_duration = (double)bytes * duration;
    double packets_duration = (double)packets * duration;

    flows++;

    packetsum += (double)packets;
    packet2sum += packets2;

    bytesum += (double)bytes;
    byte2sum += bytes2;

    dursum += (double)duration;
    dur2sum+= duration2;

    proto_flows[proto]++;
    proto_packets[proto]+=packets;
    proto_bytes[proto]+= (double) bytes;
    proto_duration[proto]+=(double) duration;

    sum2_packets[proto]+=packets2;
    sum2_bytes[proto]+=bytes2;
    sum2_duration[proto]+=duration2;
    sum_pb[proto]+=packets_bytes;
    sum_bd[proto]+=bytes_duration;
    sum_pd[proto]+=packets_duration;
    sumpb+=packets_bytes;
    sumbd+=bytes_duration;
    sumpd+=packets_duration;

#if EXTRA_TCP_STATS
    if( proto == IPPROTO_TCP )
    {
        tcp_flows++;
        tcp_bytes += (double) bytes;
        tcp_packets += packets;

        Bit64 claimed = hr->hr_maxseq - hr->hr_minseq;
        Bit64 possible = hr->hr_bytes - (hr->hr_packets * (sizeof(ip) + sizeof(tcphdr)));
        if( hr->hr_tcpflags & TH_SYN )
            possible = possible + 1;
        if( hr->hr_tcpflags & TH_FIN )
            possible = possible + 1;
        if( claimed <= possible )
            tcp_seq_bytes += claimed;
        else
        {
            tcp_seq_bytes += possible;
            claimed -= possible;
            if( claimed > BIG_BYTES_BREAKPOINT )
            {
#if DEBUG_EXTRA_TCP_STATS
                dump_flow( stdout, hr );
#endif
                tcp_seq_big_packets_leftover += packets;
                tcp_seq_big_bytes_leftover += claimed;
                tcp_seq_big_flows_leftover++;
            }
            else
            {
                tcp_seq_small_packets_leftover += packets;
                tcp_seq_small_bytes_leftover += claimed;
                tcp_seq_small_flows_leftover++;
            }
        }
        tcp_seq_n_holes[ hr->hr_fs.fs_tcp_seq_n_holes ]++;
        tcp_seq_1_holes[ hr->hr_fs.fs_tcp_seq_1_holes ]++;
        tcp_seq_hole_pkts[ hr->hr_fs.fs_tcp_seq_hole_pkts ]++;
        tcp_seq_backs[ hr->hr_fs.fs_tcp_seq_backs ]++;
        tcp_seq_dups[ hr->hr_fs.fs_tcp_seq_dups ]++;

        tcp_ack_bytes += hr->hr_maxack - hr->hr_minack;
        tcp_ack_dups[ hr->hr_fs.fs_tcp_ack_dups ]++;

        tcp_flags[ hr->hr_tcpflags ]++;
        tcp_win_max[ 1 + msb_of_native_long( hr->hr_fs.fs_tcp_win_max
#if CELLS_PER_PACKET > 1
            << (hr->hr_wscale>TCP_MAX_WINSHIFT ? TCP_MAX_WINSHIFT : hr->hr_wscale)
#endif
        ) ]++;
        tcp_bad_length[ hr->hr_fs.fs_tcp_bad_length ]++;
#if CELLS_PER_PACKET > 1
        tcp_options[ hr->hr_tcpopts ]++;
        tcp_max_segment[ hr->hr_mss ]++;
        tcp_win_scale[ hr->hr_wscale ]++;
#endif
    }
#endif

#if REASM_IP_FRAGS
    update_fragment_stats( hr, proto, packets, bytes, duration
#if KEEP_FRAG_SQUARES
        , packets2, bytes2, duration2, packets_bytes, bytes_duration, packets_duration
#endif
    );
#endif
    update_port_stats( hr, proto, packets, bytes, duration, packets2, bytes2, duration2, packets_bytes, bytes_duration, packets_duration );
    update_host_stats( hr, proto, packets, bytes, duration, packets2, bytes2, duration2, packets_bytes, bytes_duration, packets_duration );
    update_net_stats( hr, packets, bytes, duration, packets2, bytes2, duration2, packets_bytes, bytes_duration, packets_duration );
}



int report_stats( tcp_Socket *sock, float istamp, float trace_duration )
{
        int result;
        int count;
        FILE *statsfile2;
        unsigned interface;
#if REPORT_OVERHEAD_STATS
        PentiumClock start, now;
        double elapsed;
        char report_oh_line[ 200 ];
        int report_oh_so_far = 0;
        int report_oh_now;
#endif

#if !ALLOW_DISK_WRITE
        fprintf( stderr, "simulated report after %g seconds to file '%s' done\n", trace_duration, filename );
        return 0;
#endif

#if REPORT_OVERHEAD_STATS
        start.set();
#endif
        sock_printf( sock, "following statistics were calculated at query time: (now)\n" );

        sock_printf (sock, "ST\t%.0f\t%.3f\t%d\t%.f\t%.lf\t%.0f\t%.0f\t%.0f\t%.0f\t"
                                "1.0\t1.0\t1.0\t1.0\t"
                                "%.3f\t%.0f\t%.0f\t"
                                "%.3f\t%.0f\t%.3f\t%.0f\t"
                                "%.3f\t%.3f\t%.3f\n",
                istamp,
                trace_duration,

                flows,
                packetsum,
                bytesum,
                dursum,

                flows/nonzero(trace_duration),
                packetsum/nonzero(trace_duration),
                bytesum/nonzero(trace_duration),

                (float) packetsum/nonzero(flows),
                (float) bytesum/nonzero(flows),
                dursum/nonzero(flows),

                (float) bytesum/nonzero(packetsum),

                variance(flows,packetsum,packet2sum),
                variance(flows,bytesum,byte2sum),
                variance(flows,dursum,dur2sum),

                sumpb,
                sumbd,
                sumpd);

        /* this is just a legend, ignored by postprocessing
         */
        sock_printf( sock, "SP [SB] lines columns:\n"
            "   Prot [sport dport] flows packets bytes dur %%f %%p %%b %%d fps\n"
            "   pps bps ppf bpf spf bpp vp vb vd copb cobd copd\n"
        );

        if (flows) {
            for (count=0;count<=MAX_PROTOCOL;count++) {
                    if (proto_flows[count])  {
                            sock_printf(sock,"SP\t%d\t%d\t%d\t%.f\t%.f\t"
                                    "%.3f\t%.3f\t%.3f\t%.3f\t"
                                    "%.3f\t%.0f\t%.0f\t"
                                    "%.3f\t%.0f\t%.3f\t%.0f\t"
                                    "%.3f\t%.3f\t%.3f\t"
                                    "%.3f\t%.3f\t%.3f\n",
                                    count,
                                    proto_flows[count],
                                    proto_packets[count],
                                    proto_bytes[count],
                                    proto_duration[count],
                                    (float) proto_flows[count]/flows,
                                    (float) proto_packets[count]/packetsum,
                                    (float) proto_bytes[count]/bytesum,         // can bytes be zero if we got a packet?
                                    proto_duration[count]/dursum,
                                    proto_flows[count]/nonzero(trace_duration),
                                    proto_packets[count]/nonzero(trace_duration),
                                    proto_bytes[count]/nonzero(trace_duration),
                                    (float) proto_packets[count]/proto_flows[count],
                                    (float) proto_bytes[count]/proto_flows[count],
                                    proto_duration[count]/proto_flows[count],
                                    (float) proto_bytes[count]/proto_packets[count],
                                    variance(proto_flows[count],proto_packets[count],sum2_packets[count]),
                                    variance(proto_flows[count],proto_bytes[count],sum2_bytes[count]),
                                    variance(proto_flows[count],proto_duration[count],sum2_duration[count]),
                                    sum_pb[count],
                                    sum_bd[count],
                                    sum_pd[count]);
                    }
            }
        }

#if REPORT_OVERHEAD_STATS
    now.set();
    elapsed = PentiumClock( now - start ).get_seconds();
    sprintf( report_oh_line+report_oh_so_far, "ST&SP:%.4f %n", elapsed, &report_oh_now );
    report_oh_so_far += report_oh_now;
    start = now;
#endif

    // give the port-specific stats a chance to talk
    //
    result = report_port_stats( sock, trace_duration, flows, packetsum, bytesum, dursum );
    if( result )
        return result;

#if REPORT_OVERHEAD_STATS
    now.set();
    elapsed = PentiumClock( now - start ).get_seconds();
    sprintf( report_oh_line+report_oh_so_far, "ports:%.4f %n", elapsed, &report_oh_now );
    report_oh_so_far += report_oh_now;
    start = now;
#endif

    // give the host-specific stats a chance to talk
    //
    result = report_host_stats( sock, trace_duration, flows, packetsum, bytesum, dursum );
    if( result )
        return result;

#if REPORT_OVERHEAD_STATS
    now.set();
    elapsed = PentiumClock( now - start ).get_seconds();
    sprintf( report_oh_line+report_oh_so_far, "hosts:%.4f %n", elapsed, &report_oh_now );
    report_oh_so_far += report_oh_now;
    start = now;
#endif

    // give the net-specific stats a chance to talk
    //
    result = report_net_stats( sock,
        istamp, trace_duration, flows,
        packetsum, bytesum, dursum,
        packet2sum, byte2sum, dur2sum,
        sumpb, sumbd, sumpd
    );
    if( result )
        return result;

#if REPORT_OVERHEAD_STATS
    now.set();
    elapsed = PentiumClock( now - start ).get_seconds();
    sprintf( report_oh_line+report_oh_so_far, "nets:%.4f\n%n", elapsed, &report_oh_now );
    report_oh_so_far += report_oh_now;
    start = now;
#endif

#if REASM_IP_FRAGS
    // give the ip-fragment-specific stats a chance to talk
    //
    result = report_fragment_stats( sock, trace_duration, flows, packetsum, bytesum, dursum );
    if( result )
        return result;

#if REPORT_OVERHEAD_STATS
    now.set();
    elapsed = PentiumClock( now - start ).get_seconds();
    sprintf( report_oh_line+report_oh_so_far, "fragments:%.4f\n%n", elapsed, &report_oh_now );
    report_oh_so_far += report_oh_now;
    start = now;
#endif
#endif

#if EXTRA_TCP_STATS
    // spew the global TCP-specific stats
    //
    sock_printf( sock, "TCP flows: %u\n", tcp_flows );
    sock_printf( sock, "TCP packets: %.0f\n", tcp_packets );
    sock_printf( sock, "TCP raw forward bytes (including IP): %.0f\n", tcp_bytes );
    sock_printf( sock, "TCP good forward bytes (not including IP or TCP or options): %.0f\n", tcp_seq_bytes );
    sock_printf( sock, "TCP good backward bytes (not including IP or TCP or options): %.0f\n", tcp_ack_bytes );
    sock_printf( sock, "TCP small leftover hole flows: %u\n", tcp_seq_small_flows_leftover );
    sock_printf( sock, "TCP small leftover hole bytes: %.0f\n", tcp_seq_small_bytes_leftover );
    sock_printf( sock, "TCP small leftover hole packets: %.0f\n", tcp_seq_small_packets_leftover );
    sock_printf( sock, "TCP big leftover hole flows: %u\n", tcp_seq_big_flows_leftover );
    sock_printf( sock, "TCP big leftover hole bytes: %.0f\n", tcp_seq_big_bytes_leftover );
    sock_printf( sock, "TCP big leftover hole packets: %.0f\n", tcp_seq_big_packets_leftover );

    sock_printf( sock, "TCP count of flows for each multiple hole count:" );
    for( count=0; count<=BIT8_MAX; count++ )
        if( tcp_seq_n_holes[ count ] )
            sock_printf( sock, "\t%d:%u", count, tcp_seq_n_holes[count] );

    sock_printf( sock, "\nTCP count of flows for each single hole count:" );
    for( count=0; count<=BIT8_MAX; count++ )
        if( tcp_seq_1_holes[ count ] )
            sock_printf( sock, "\t%d:%u", count, tcp_seq_1_holes[count] );

    sock_printf( sock, "\nTCP count of flows for each count of packets in single holes:" );
    for( count=0; count<=BIT8_MAX; count++ )
        if( tcp_seq_hole_pkts[ count ] )
            sock_printf( sock, "\t%d:%u", count, tcp_seq_hole_pkts[count] );

    sock_printf( sock, "\nTCP count of flows for each count of backtracking packets:" );
    for( count=0; count<=BIT8_MAX; count++ )
        if( tcp_seq_backs[ count ] )
            sock_printf( sock, "\t%d:%u", count, tcp_seq_backs[count] );

    sock_printf( sock, "\nTCP count of flows for each count of packets with duplicate forward bytes:" );
    for( count=0; count<=BIT8_MAX; count++ )
        if( tcp_seq_dups[ count ] )
            sock_printf( sock, "\t%d:%u", count, tcp_seq_dups[count] );

    sock_printf( sock, "\nTCP count of flows for each count of packets with duplicate backward bytes:" );
    for( count=0; count<=BIT8_MAX; count++ )
        if( tcp_ack_dups[ count ] )
            sock_printf( sock, "\t%d:%u", count, tcp_ack_dups[count] );

    sock_printf( sock, "\nTCP count of flows for each count of packets with bad TCP header length:" );
    for( count=0; count<=BIT8_MAX; count++ )
        if( tcp_bad_length[ count ] )
            sock_printf( sock, "\t%d:%u", count, tcp_bad_length[count] );

    sock_printf( sock, "\nTCP count of flows for each combination of TCP flags:" );
    for( count=0; count<=BIT8_MAX; count++ )
        if( tcp_flags[ count ] )
            sock_printf( sock, "\t(%d)%s:%u", count, tcp_flags_string[count], tcp_flags[count] );

#if CELLS_PER_PACKET > 1
    sock_printf( sock, "\nTCP count of flows for each combination of TCP options:" );
    for( count=0; count<TCP_OPTION_MAX; count++ )
        if( tcp_options[ count ] )
            sock_printf( sock, "\t(%d)%s:%u", count, tcp_options_string[count], tcp_options[count] );
#endif

    sock_printf( sock, "\nTCP count of flows for each log2 of max RX window size:" );
    for( count=0; count<=BIT32_LOG2_MAX; count++ )
        if( tcp_win_max[ count ] )
            sock_printf( sock, "\t%d:%u", count-1, tcp_win_max[count] );

#if CELLS_PER_PACKET > 1
    sock_printf( sock, "\nTCP count of flows for each window shift count option:" );
    for( count=0; count<TCP_MAX_WINSHIFT+2; count++ )
        if( tcp_win_scale[ count ] )
            sock_printf( sock, "\t%d:%u", count, tcp_win_scale[count] );
#endif

#if CELLS_PER_PACKET > 1
    sock_printf( sock, "\nTCP count of flows for each maximum segment size option:" );
    for( count=0; count<BIT16_MAX; count++ )
        if( tcp_max_segment[ count ] )
            sock_printf( sock, "\t%d:%u", count, tcp_max_segment[count] );
#endif

    sock_printf( sock, "\n" );
#if REPORT_OVERHEAD_STATS
    now.set();
    elapsed = PentiumClock( now - start ).get_seconds();
    sprintf( report_oh_line+report_oh_so_far, "tcp_extra:%.4f\n%n", elapsed, &report_oh_now );
    report_oh_so_far += report_oh_now;
    start = now;
#endif
#endif

#if REPORT_OVERHEAD_STATS
    fputs( report_oh_line, stderr );
    sock_puts( sock, report_oh_line );
#endif

#if DEBUG_VARIANCE
    float ftemp;

    ftemp = variance(flows,packetsum,packet2sum);
    if( ftemp < 0 )
        sock_printf( sock, "negative packet/flow variance %f caused by flows=%u packetsum=%f packet2sum=%f\n", ftemp, flows, packetsum, packet2sum );

    ftemp = variance(flows,bytesum,byte2sum);
    if( ftemp < 0 )
        sock_printf( sock, "negative byte/flow variance %f caused by flows=%u bytesum=%f byte2sum=%f\n", ftemp, flows, bytesum, byte2sum );

    ftemp = variance(flows,dursum,dur2sum);
    if( ftemp < 0 )
        sock_printf( sock, "negative duration/flow variance %f caused by flows=%u dursum=%f dur2sum=%f\n", ftemp, flows, dursum, dur2sum );
#endif

#if FORCE_LARGE_FILES
    // append 100K of trailing gunk to the text report
    // to see if capture and/or flows and/or TCP breaks
    //
    for( i=0; i<2000; i++ )

        // 50 bytes each line
        //
        sock_printf( sock, "123456789012345678901234567890123456789012345678\n" );
#endif

    // ees no my yob to close socket!
    //

    return 0;
}



#endif  // DO_FOREGROUND_STATS
