/* conditional compile switches all go here
 */

#define BYTESWAP_FIELDS_EARLY   FALSE       /* must match fs2flows.c */
#define USE_OLD_HASH            FALSE
#define ALWAYS_DUMP_HASH_INFO   FALSE
#define FOLD_BEFORE_MODULUS     FALSE
#define FOLD_RANDOMIZES         TRUE
#define FOLD_KEEPS_16_BITS      TRUE
#define DEBUG_CHAIN_LENGTH      FALSE
#define DEBUG_HASHINDEX         FALSE
#define DEBUG_WORKING_SET       TRUE
#define DEBUG                   FALSE
#define SORT_PORT_PRINTOUT      FALSE
#define TALK_WHEN_ALLOC_FAILS   FALSE           // for debugging when overflow occurs only!  too slow




#define MAX_PROTOCOL 255
#define MAX_PORT 65536
#define PORTREF_POOL_SIZE       ((MAX_PORT+1) * sizeof(int))

#if __unix__
#   include <sys/types.h>
#   include <sys/time.h>
#   include <netinet/in.h>
#else
#   include "../wattcp/tcp.h"
#endif
#include <stdio.h>
#include <stdlib.h>             /* errno */
#include <fcntl.h>

#if __unix__
#   include "types.h"
#else
#   include "../types/types.h"
#   include "../intel/byteswap.h"
#endif

#include "queue.h"
#include "trace.h"              /* must be included before util.h */
#include "util.h"
#include "dostats.h"            /* nonzero() variance() */
#include "portstat.h"
#include "../dpmi32/flat.h"     /* flat_alloc() */



// none of this code gets compiled in
// if it's isn't going to be called
//
#if DO_FOREGROUND_STATS



static int      *port_ref = NULL;
static int      portthreshold;
static float    cutoffp, cutoffb, cutofff;
static Bit32    tportct, uportct;
static int      portpaircount;

#if USE_OLD_HASH
#define N_PORTQUADS 4096

/*
 * port quad hash function
 *  next 4 bits: low 4 bits of IP prot #
 *  next 4 bits: low 4 bits of src port #
 *  next 4 bits: low 4 bits of dst port #
 *
 * assumption is that the 1MB of address range given this configuration
 * is large enough to avoid a high table load factor
 */

#define porthash(prot,sport,dport) ( (((u_int)prot)&0xf << 8) | \
                  (((u_int)sport)&0xf << 4) | \
                  (((u_int)dport)&0xf) )
#else
/*
#define N_PORTQUADS    104729
#define N_PORTQUADS    32749
#define N_PORTQUADS    8191
#define N_PORTQUADS    4001
#define N_PORTQUADS    1021
#define N_PORTQUADS    251
*/
#define N_PORTQUADS    65536

#if !FOLD_BEFORE_MODULUS
#   define fold(x) (x)
#   define modulus(x) ((x) % N_PORTQUADS)
#else
#if FOLD_RANDOMIZES
#   if BIT64_AVAILABLE
#       define fold32(x) (((x)>>32) ^ (x))
#       define modulus(x) ((x) & 65535)
#   else
#       define fold32(x) (x)
#       define modulus(x) (x)
#   endif
#   define fold(x) ((fold32(x)*1664525 + 1013904223) >> 16)
#else
#if FOLD_KEEPS_16_BITS
#   define fold(x) (((x)>>16) ^ (x))
#   define modulus(x) ((x) & 65535)
#else
#   define fold(x) (((x)>>24) ^ ((x)>>16) ^ ((x)>>8) ^ (x))
#   define modulus(x) ((x) & 255)
#endif
#endif
#endif

/* generate a 32-bit hash key
 */
#define porthash(prot,sport,dport) modulus( fold( \
                  (prot) + (sport) + (dport) \
                  ) )
#endif      /* use new hash */

static struct porthashrec **porttabfull = NULL;
static struct porthashrec **porttab = NULL;
#define PORTBUCKET_POOL_SIZE (N_PORTQUADS * sizeof(struct porthashrec *))

static struct porthashrec *porthash_free = NULL;
static struct porthashrec *porthash_pool = NULL;
u_int porthash_alloc_failures;

// 15000L was too small 4:20PM 7/15/96 WTN
// 20000L was too small 4:00PM 8/18/96 WTN
// 30000L was too small 4:35PM 10/22/96 WTN
#define PORTHASH_RECORDS 60000L

#define PORTHASH_POOL_SIZE (PORTHASH_RECORDS * sizeof(struct porthashrec))

#if DEBUG_CHAIN_LENGTH
#   define PORTCHAIN_POOL_SIZE (N_PORTQUADS * sizeof(int))
    static int *chainlen2 = NULL;
    static int *chainlen3 = NULL;
    static int longest_chain2;
    static int longest_chain3;
#endif

#if DEBUG_WORKING_SET
    static int total_level2_porthashrecs;
    static int total_level3_porthashrecs;
#endif

#if DEBUG_HASHINDEX
    static u_int hashindex_trace = -1;
#endif



#if SORT_PORT_PRINTOUT
#define PORTPRINT_POOL_SIZE (PORTHASH_RECORDS * sizeof(struct porthashrec *))
static Bit32 portcount;
static struct porthashrec **portprintout = NULL;

static int portcmp( void *a, void *b )
{
    struct porthashrec ** hrp1 = (struct porthashrec**) a;
    struct porthashrec ** hrp2 = (struct porthashrec**) b;
    return ((*hrp1)->packets - (*hrp2)->packets);
}
#endif



/*
 * new_port_record
 *      establish a new port triple
 *      1> insert into hash table [maintained as stack]
 */
static
struct porthashrec *
new_port_record( u_int prot, u_int sport, u_int dport,
    u_int flows, u_int packets, Bit64 bytes, float duration,
    double p2sum, double b2sum, double d2sum,
    double pbsum, double bdsum, double pdsum
)
{
    struct porthashrec *pr;

    if ((pr = porthash_free) == NULL) {
#if TALK_WHEN_ALLOC_FAILS
        fprintf(stderr, "unable to alloc new porthashrec\n");
#endif
        porthash_alloc_failures++;
        return 0;
    }
    porthash_free = pr->next;

    pr->proto=prot;
    pr->sport=sport;
    pr->dport=dport;
    pr->flows=flows;
    pr->packets=packets;
    pr->bytes=bytes;
    pr->duration=duration;
    pr->p2sum=p2sum;
    pr->b2sum=b2sum;
    pr->d2sum=d2sum;
    pr->pbsum=pbsum;
    pr->bdsum=bdsum;
    pr->pdsum=pdsum;
    pr->next=NULL;
    return(pr);
}



static void delete_port_record( struct porthashrec *pr )
{
    /* put this puppy on the free list */
    pr->next = porthash_free;
    porthash_free = pr;
}


/*
 * diagnostic
 */
static void print_port_record( struct porthashrec *pr )
{
    fprintf(stderr, "p: %d ", pr->packets);
    fprintf(stderr, "b: %.0f ", (double)pr->bytes);
    fprintf(stderr, "psdd hash: %d %d %d %f %d\n",
        pr->proto,pr->sport,pr->dport,pr->duration,
        porthash(pr->proto,pr->sport,pr->dport));
}



// forgets all statistics
//
int init_port_stats( void )
{
    Bit32 count;

    cutoffp = .0005;
    cutoffb = .0005;
    cutofff = .0005;
    portthreshold = 5;

    tportct = 0;
    uportct = 0;
    portpaircount = 0;

#if DEBUG_CHAIN_LENGTH
    longest_chain2 = 0;
    longest_chain3 = 0;
#endif

#if DEBUG_WORKING_SET
    total_level2_porthashrecs = 0;
    total_level3_porthashrecs = 0;
#endif

    // if this is the first clear (really initialization)
    //
    if( porthash_pool == NULL )
    {
        fprintf( stderr, "port hash records are %d bytes each\n", sizeof(struct porthashrec) );

#if DEBUG_CHAIN_LENGTH
        /* allocate chain length counters for each bucket in main hash table */
        chainlen2 = (int *) flat_alloc( PORTCHAIN_POOL_SIZE );
        if( chainlen2 == NULL )
        {
            printf( "error %d allocating %d bytes (%d integers) for main port chain length counters\n", errno, PORTCHAIN_POOL_SIZE, N_PORTQUADS );
            return 1;
        }

        /* allocate chain length counters for each bucket in secondary hash table */
        chainlen3 = (int *) flat_alloc( PORTCHAIN_POOL_SIZE );
        if( chainlen3 == NULL )
        {
            printf( "error %d allocating %d bytes (%d integers) for secondary port chain length counters\n", errno, PORTCHAIN_POOL_SIZE, N_PORTQUADS );
            return 1;
        }
#endif

        // allocate the port reference counters
        //
        port_ref = (int *) flat_alloc( PORTREF_POOL_SIZE );
        if( port_ref == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld integers) for port reference counters\n", errno, PORTREF_POOL_SIZE, MAX_PORT+1 );
            return 1;
        }

        /* allocate the buckets for the main hash table */
        porttabfull = (struct porthashrec **) flat_alloc( PORTBUCKET_POOL_SIZE );
        if( porttabfull == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld pointers) for main port hash buckets\n", errno, PORTBUCKET_POOL_SIZE, N_PORTQUADS );
            return 1;
        }

        /* allocate the buckets for the secondary hash table */
        porttab = (struct porthashrec **) flat_alloc( PORTBUCKET_POOL_SIZE );
        if( porttab == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld pointers) for secondary port hash buckets\n", errno, PORTBUCKET_POOL_SIZE, N_PORTQUADS );
            return 1;
        }

#if SORT_PORT_PRINTOUT
        /* allocate pointers to sort during report */
        portprintout = (struct porthashrec **) flat_alloc( PORTPRINT_POOL_SIZE );
        if( portprintout == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld pointers) for port printout sorting\n", errno, PORTPRINT_POOL_SIZE, PORTHASH_RECORDS );
            return 1;
        }
#endif

        /* allocate a pool of hash records */
        porthash_pool = (struct porthashrec *) flat_alloc( PORTHASH_POOL_SIZE );
        if( porthash_pool == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld records) for port hash record pool\n", errno, PORTHASH_POOL_SIZE, PORTHASH_RECORDS );
            return 1;
        }

        /* put all the hash records on the singly-linked free list */
        porthash_free = porthash_pool;
        for( count=0; count<PORTHASH_RECORDS-1; count++ )
        {
            porthash_pool[ count ].next = porthash_pool + count + 1;
        }
        porthash_pool[count].next = NULL;

    }
    porthash_alloc_failures = 0;

    // no ports have been referenced yet
    //
    for (count=0;count<=MAX_PORT;count++)
    {
        port_ref[count]=0;
    }

    for( count=0; count<N_PORTQUADS; count++ )
    {
        // level2 hashtable is empty
        //
        porttabfull[count] = 0;

        // level3 hashtable is empty
        //
        porttab[count] = 0;

#if SORT_PORT_PRINTOUT
        // no port pairs are ready to print
        //
        portprintout[ count ] = 0;
#endif

#if DEBUG_CHAIN_LENGTH
        /* initialize chain length of all hashtable buckets to zero
         */
        chainlen2[ count ] = 0;
        chainlen3[ count ] = 0;
#endif
    }

    return 0;
}



void update_port_stats( struct hashrec *hr, u_char proto,
    u_int packets, Bit64 bytes, float duration,
    double packets2, double bytes2, double duration2,
    double packets_bytes, double bytes_duration, double packets_duration
)
{
        int porthashindex, cmp;
        struct porthashrec *pp, *porth, *newpp;
#if BYTESWAP_FIELDS_EARLY
        u_int  sport = hr->hr_fd.fd_sport, dport = hr->hr_fd.fd_dport;
#else
        u_int  sport = ntohs(hr->hr_fd.fd_sport), dport = ntohs(hr->hr_fd.fd_dport);
#endif

        if ( (proto==IPPROTO_TCP) || (proto==IPPROTO_UDP)) {
                if ((sport < 1000)  && (dport >= 1000))
                        { dport =0; port_ref[sport]++;  } else
                if ((dport < 1000)  && (sport >= 1000))
                        { sport =0; port_ref[dport]++;  } else
                {
                        port_ref[sport]++;
                        if (sport!=dport) port_ref[dport]++;
                }
        }
        porthashindex = porthash(proto,sport,dport);
        pp = porttabfull[porthashindex];
        /* look for match or collision on hash chain */
        porth=pp;
        cmp=0;
        while (porth!=NULL && cmp==0)
        {
                cmp = ( (proto==porth->proto) && (sport==porth->sport) &&
                                        (dport==porth->dport) );
                if (cmp) {
                        /* update all the stats in porthashrec */
                        porth->flows++;
                        porth->packets+=packets;
                        porth->bytes+=bytes;
                        porth->duration+=duration;
                        porth->p2sum+=packets2;
                        porth->b2sum+=bytes2;
                        porth->d2sum+=duration2;
                        porth->pbsum+=packets_bytes;
                        porth->bdsum+=bytes_duration;
                        porth->pdsum+=packets_duration;
#if DEBUG
                        /* for debugging */
                        printf("match\t");
#endif
                }
                porth=porth->next;
        };
        /* didn't find it anywhere, must be a new record */
        if (cmp==0)
        {
            portpaircount++;
            if( proto==IPPROTO_TCP )
                tportct++;
            else if( proto==IPPROTO_UDP )
                uportct++;
            porth = new_port_record( proto, sport, dport, 1, packets,
                bytes, duration, packets2, bytes2, duration2, packets_bytes,
                bytes_duration, packets_duration );
            if( porth )
            {
                porth->next = pp;
                porttabfull[porthashindex] = porth;
#if DEBUG_WORKING_SET
                /* increment count of level2 hash records
                 */
                total_level2_porthashrecs++;
#endif

#if DEBUG_CHAIN_LENGTH
                /*
                 * increment chain length for this hash index
                 */
                chainlen2[porthashindex]++;

#if DEBUG_HASHINDEX
                if( porthashindex == hashindex_trace )
                    fprintf( stderr, "at packet %d hash index 0x%8.8lX chain length increased to %d\n", total_packets, porthashindex, chainlen2[porthashindex] );
#endif

                /*
                 * if this is the longest chain we have ever seen,
                 * update our global, and notify the user
                 *
                 * notifications should be asymptotically decreasing
                 */
                if( longest_chain2 < chainlen2[porthashindex] )
                {
                    longest_chain2 = chainlen2[porthashindex];
                    fprintf( stderr, "packet %d hash index 0x%8.8lX is now longest port chain 2 at %d\n", total_packets, porthashindex, longest_chain2 );
                    fprintf( stderr, "calced from proto=%d TCP=%4.4X:%4.4X\n", proto, sport, dport );
                }
#endif

/* for debugging */
#if DEBUG
                printf("new\t");
#endif
            }   // end if could alloc porthashrec
        }   // end if needed a new porthashrec
#if DEBUG
        print_port_record(porth);
#endif
}



/* "SB proto sport dport flows packets bytes dur %%f %%p %%b %%d fps "
   "pps bps ppf bpf spf bpp vp vb vd copb cobd copd\n");
*/
static void print_port_stats( tcp_Socket *sock, float trace_duration,
    int total_flows, struct porthashrec *porth, double packetsum,
    double bytesum, double dursum )
{
    sock_printf(
        sock,
        "SB\t%d\t%d\t%d\t%d\t%d\t%.0f\t%.3f\t"
            "%.3f\t%.3f\t%.3f\t%.3f\t"
            "%.3f\t%.0f\t%.0f\t"
            "%.3f\t%.0f\t%.3f\t%.0f\t"
            "%.3f\t%.3f\t%.3f\t"
            "%.3f\t%.3f\t%.3f\n",

        porth->proto,
        porth->sport,
        porth->dport,

        porth->flows,
        porth->packets,
        (double) porth->bytes,
        porth->duration,

        (float) porth->flows/total_flows,
        (float) porth->packets/packetsum,
        (float) porth->bytes/bytesum,
        (float) porth->duration/dursum,

        porth->flows/nonzero(trace_duration),
        porth->packets/nonzero(trace_duration),
        porth->bytes/nonzero(trace_duration),

        (float) porth->packets/porth->flows,
        (float) porth->bytes/porth->flows,
        (float) porth->duration/porth->flows,

        (float) porth->bytes/porth->packets,

        variance(porth->flows,porth->packets,porth->p2sum),
        variance(porth->flows,porth->bytes,porth->b2sum),
        variance(porth->flows,porth->duration,porth->d2sum),

        (float) porth->pbsum,
        (float) porth->bdsum,
        (float) porth->pdsum                // 24th field
    );
}



int report_port_stats( tcp_Socket *sock, float trace_duration, u_int flows,
    double packetsum, double bytesum, double dursum )
{
        Bit64 cutoffb_bit64;
        int i, expcount=0, exprefcount=0;
        int porthashindex, cmp;
        struct porthashrec *pp, *porthfull, *porth, *newpp, *next;

#if DEBUG_WORKING_SET
        fprintf( stderr, "%d level 2 port hash records created\n", total_level2_porthashrecs );
#endif

        /* Display ports with excessive references in pairings.  */
        /*      # SS port1:ref1 port2:ref2 port3:ref3 */
        sock_printf(sock,"SS\t");
        for (i=0;i<=MAX_PORT;i++) {
            if (port_ref[i] >= portthreshold) {
                /* sock_printf(sock,"port %5d is found paired %6d times; exp %d exp2 %d .\n", */
                sock_printf(sock,"%d:%d\t",i,port_ref[i]);
                expcount++;
                exprefcount = exprefcount + port_ref[i];
            }
        };
        sock_printf(sock,"\n");

        /* SC #ports #refs #TCP<1000 #UDP<1000 #ports<1000       */
        sock_printf(sock,"SC\t%d\t%d\t%d\t%d\t%d\t%d\n",
                portthreshold,expcount,exprefcount,tportct,uportct,portpaircount);

        /* collapse further
         *
         * we could not do this on the first pass because
         * we did not know (total usage per port) yet
         */
        for (i=0;i<N_PORTQUADS;i++) {
            for (porthfull=porttabfull[i]; porthfull!=NULL; porthfull=next)
            {
                next = porthfull->next;

                // if protocol is either TCP or UDP...
                //
                if ( (porthfull->proto==IPPROTO_TCP) || (porthfull->proto==IPPROTO_UDP))
                {
                    // if source port had more than the threshold flows
                    // and source port is not already collapsed
                    // and source port had more flows than dest port...
                    //
                    if ((port_ref[porthfull->sport] > portthreshold) &&
                        (porthfull->sport != 0) &&
                        (port_ref[porthfull->dport] < port_ref[porthfull->sport]))
                    {
                        porthfull->dport = 0;
                    }
                    else
                    {
                        if ((port_ref[porthfull->dport] > portthreshold) &&
                            (porthfull->dport != 0) &&
                            (port_ref[porthfull->sport] < port_ref[porthfull->dport]))
                        {
                            porthfull->sport = 0;
                        }
                    }
                }   /* end if protocol is TCP or UDP */

                /* rehash into new array */
                porthashindex = porthash(porthfull->proto,porthfull->sport,
                                        porthfull->dport);
                pp = porttab[porthashindex];
                porth=pp;
                cmp=0;
                while (porth != NULL && cmp==0) {
/* for debugging */
#if DEBUG
                    fprintf(stderr,"seeking match portab\t");
                    print_port_record(porth);
#endif
                    cmp = ( (porthfull->proto==porth->proto) &&
                                (porthfull->sport==porth->sport) &&
                                (porthfull->dport==porth->dport) );
                    if (cmp) {
                        porth->flows+=porthfull->flows;
                        porth->packets+=porthfull->packets;
                        porth->bytes+=porthfull->bytes;
                        porth->duration+=porthfull->duration;
                        porth->p2sum+=porthfull->p2sum;
                        porth->b2sum+=porthfull->b2sum;
                        porth->d2sum+=porthfull->d2sum;
                        porth->pbsum+=porthfull->pbsum;
                        porth->bdsum+=porthfull->bdsum;
                        porth->pdsum+=porthfull->pdsum;
    /* for debugging */
#if DEBUG
                        fprintf(stderr,"match portab\t");
                        print_port_record(porth);
#endif
                    }
                    porth=porth->next;

                }   /* end search hash chain 3 */

/* didn't find it anywhere, must be a new record */
                if (cmp == 0)
                {
                    newpp = new_port_record(
                        porthfull->proto, porthfull->sport, porthfull->dport,
                        porthfull->flows,
                        porthfull->packets, porthfull->bytes, porthfull->duration,
                        porthfull->p2sum, porthfull->b2sum, porthfull->d2sum,
                        porthfull->pbsum, porthfull->bdsum, porthfull->pdsum );
                    if( newpp )
                    {
                        newpp->next = pp;
                        porttab[porthashindex] = newpp;
#if DEBUG_WORKING_SET
                        /* increment count of level3 hash records
                         */
                        total_level3_porthashrecs++;
#endif
#if DEBUG_CHAIN_LENGTH
                        /*
                         * increment chain length for this hash index
                         */
                        chainlen3[porthashindex]++;

                        /*
                         * if this is the longest chain we have ever seen,
                         * update our global, and notify the user
                         *
                         * notifications should be asymptotically decreasing
                         */
                        if( longest_chain3 < chainlen3[porthashindex] )
                        {
                            longest_chain3 = chainlen3[porthashindex];
                            fprintf( stderr, "packet %d hash index 0x%8.8lX is now longest port chain 3 at %d\n", total_packets, porthashindex, longest_chain3 );
                            fprintf( stderr, "calced from proto=%d TCP=%4.4X:%4.4X\n", porthfull->proto, porthfull->sport, porthfull->dport );
                        }
#endif
                    }   // end if could alloc porthashrec
                }

                /* free up the porttabfull entry (level 2 hash)
                 *
                 * JZA 2/5/96   need to re-enable this
                 * so code can run continuously...
                 */
                delete_port_record( porthfull );

            }   /* end search hash chain 2 */
        }   /* end for all level 2 chains */

#if DEBUG_WORKING_SET
        fprintf( stderr, "%d level 3 port hash records created\n", total_level3_porthashrecs );
#endif

        /* go through porttab and print now */
        cutoffp *= packetsum;
        cutoffb *= bytesum;
        cutoffb_bit64 = cutoffb;
        cutofff *= flows;

        for (i=0;i<N_PORTQUADS;i++) {
            for (porth=porttab[i]; porth!=NULL; porth=next) {
                next = porth->next;

                if ((porth->packets > cutoffp) ||
                    (porth->bytes > cutoffb_bit64) ||
                    (porth->flows > cutofff))
                {
#if SORT_PORT_PRINTOUT
                    /* portprintout: array of pointers to struct porthashrec's */
                    portcount++;
                    portprintout[portcount]=porth;
#else
                    print_port_stats(sock, trace_duration, flows, porth, packetsum, bytesum, dursum);
#endif
                }
#if SORT_PORT_PRINTOUT
                /* only want to delete unselected port pairs now;
                 * selected ones will be deleted after sort/print
                 */
                else
#else
                /* want to delete both printed and unprinted port pairs
                 */
#endif
                    delete_port_record( porth );
            }
        }

#if SORT_PORT_PRINTOUT
        /* sort from porthle array 80 each */
        qsort( portprintout, portcount, sizeof(struct porthashrec *), &portcmp );
        for( i=0; i<portcount; i++)
        {
            print_port_stats( sock, trace_duration, portprintout[i],packetsum, bytesum, dursum );
            delete_port_record( portprintout[i] );
        }
#endif

    return 0;
}



#endif  // DO_FOREGROUND_STATS
