#include <stdio.h>                      // fopen()
#include <errno.h>                      // errno
#include <conio.h>                      // kbhit()
#include <mem.h>                        // memcpy()
#include "trace.h"
#include "..\intel\byteswap.h"          // htons() htonl()



// conditional compilation switches
//
#define TRACE_ALREADY_SWAPPED           TRUE



// global variables
//
struct traceblock in_buffer;
struct traceblock_send out_buffer;



int main( int argc, char *argv[] )
{
    const char *program = argv[ 0 ];
    const char *infile  = argv[ 1 ];
    const char *outfile = argv[ 2 ];
    FILE *in_handle;
    FILE *out_handle;
    int result;
    Bit32 used;
    Bit32 count;
    Bit32 temp;
    unsigned long block_count = 0;

    printf( "each captured record is %d bytes\n", sizeof(struct tracerec) );
    printf( "each sendable record is %d bytes\n", sizeof(struct tracerec_send) );
    printf( "%d captured records in a block\n", COMPL_RING_ENTRIES );
    printf( "%d sendable records in a block\n", SEND_RING_ENTRIES );
    printf( "\n" );

    if( argc != 3 )
    {
        printf( "converts an OCxMON raw capture file to a sendable file\n\n" );
        printf( "usage: %s infile outfile\n", program );
        return 1;
    }

    in_handle = fopen( infile, "rb" );
    if( in_handle == NULL )
    {
        printf( "error %d opening input file\n", errno );
        return 2;
    }

    out_handle = fopen( outfile, "wb" );
    if( out_handle == NULL )
    {
        printf( "error %d opening output file\n", errno );
        return 3;
    }

    while( !feof( in_handle ) && !ferror( in_handle ) && !kbhit() )
    {
        result = fread( &in_buffer, sizeof(in_buffer), 1, in_handle );
        if( result != 1 )
        {
            printf( "error %d reading input file\n", errno );
            break;
        }

#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
        // must remember usage count before swapping
        // when we swap on the Intel machine
        //
        used = ntohl( in_buffer.used );
#endif

        // copy the card index and (big-endian) usage count
        //
        // note: since send records are smaller than captured
        // records, more send records fit into a send block
        // than were originally in the capture block,
        // so we will be wasting some space at the end of
        // the send block.  oh well.
        //
        out_buffer.cardindex = in_buffer.cardindex;
        out_buffer.used = in_buffer.used;

        // check for out-of-bound record count
        //
        if( used > COMPL_RING_ENTRIES )
        {
            printf( "error: usage count of block %u is %u, but max is %u!\n", block_count, used, COMPL_RING_ENTRIES );
            break;
        }

        // for each record in the buffer...
        //
        for( count=0; count<used; count++ )
        {
            // no need to swap the timestamp since we won't be keeping it
            //

            // repeat count will always be one
            //
            out_buffer.data[count].count = 1;

#if CARD_SWAPS_ATM_HEADER
            // swap the bytes of the ATM header
            //
            out_buffer.data[count].atm.whole = htonl( in_buffer.data[count].atm.whole );
#else
            out_buffer.data[count].atm.whole = in_buffer.data[count].atm.whole;
#endif

            // the payload was always in network byte order
            //
            memcpy( &out_buffer.data[count].padding[0], (1 + &in_buffer.data[count].atm), CELL_SIZE );
        }

        result = fwrite( &out_buffer, sizeof(out_buffer), 1, out_handle );
        if( result != 1 )
        {
            printf( "error %d writing output file\n", errno );
            break;
        }

        block_count++;
        printf( "%d(%d) ", block_count, in_buffer.cardindex );
    }
    printf( "\n" );

    if( kbhit() )
    {
        getch();
        fprintf( stderr, "halted by keypress!\n" );
    }

    fclose( out_handle );
    fclose( in_handle );

    return 0;
}

