// this source file should be linked with CAPTURE.CPP
// instead of FS2FLOWS.C and DOSTATS.C and
// PORTSTAT.C and NETSTATS.C and HOSTSTAT.C and
// UTIL.C and READBGP.C and SUBNET.C and QUEUE.C
//



#include <string.h>                         // strcpy()
#include "../integrat/buffer.h"             // class BufferBlock class BuffferQueue
#include "../integrat/fs2flows.h"           // prototypes we must adhere to



// manifest constants
//
#define MAX_FILENAME_SIZE       1000



// global variables
//
int want_disk_io_during_run = FALSE;
int file_is_open = FALSE;
char filename[ MAX_FILENAME_SIZE ];
unsigned block_count = 0;
BufferBlock *head = NULL;
BufferBlock *tail = NULL;



// note: flows_per_block() is not implemented
// because AAL5.C function cells_reassembly() won't be called
//



int flows_per_block( BufferBlock *block )
{
    int result;

    if( !file_is_open )
    {
        // try to open the file
        //
        result = all_buffers->dump_open( filename );
        if( result )
        {
            fprintf( stderr, "error %d from BufferQueue::dump_open('%s')\n", result, filename );
            return result;
        }

        fprintf( stderr, "BufferBlock's will be saved to file '%s' %s\n",
            filename,
            want_disk_io_during_run ? "now" : "later"
        );

        file_is_open = TRUE;
    }

    if( want_disk_io_during_run )
    {
        // write this block to disk now
        //
        result = all_buffers->dump_one( block );
        if( result )
        {
            fprintf( stderr, "error %d from BufferQueue::dump_one()\n", result );
            return result;
        }

        block_count++;
#if BUFFER_DUMP_TICKS
        fprintf( stderr, "%4d\b\b\b\b", block_count );
#endif

        // free this block now
        //
        all_buffers->free( block );
    }
    else
    {
        // link this block onto our own list
        //
        block->next = NULL;
        if( tail )
            tail->next = block;
        else
            head = block;
        tail = block;
    }

    return 0;
}



// turn off compiler warning about unused parameters
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
#pragma warn -par

int flows_init( int argc, char *argv[], PentiumClock *keystroke_interval,
    PentiumClock *query_interval, PentiumClock *active_interval,
    int *want_facility_loopback, int *want_rx, int *want_tx )
{
    int result;
    const char *program = argv[ 0 ];
    const char *filename0 = argv[ 1 ];

    if( argc!=2 && argc!=3 )
    {
        fprintf( stderr, "\n" );
        fprintf( stderr, "usage: %s filename {placeholder}\n", program );
        fprintf( stderr, "anything in second parm causes disk write during capture, instead of after\n" );
        fprintf( stderr, "\n" );
        return 1;
    }

    // we cannot call any BufferQueue member functions
    // because the global pointer "all_buffers" has not
    // been assigned to yet
    //
    // so remember the filename and open the file later
    // the first time we need it
    //
    strcpy( filename, filename0 );
    file_is_open = FALSE;

    if( argc == 3 )
        want_disk_io_during_run = TRUE;

    if( want_disk_io_during_run )
        fprintf( stderr, "\ncapturing cells to disk until it is full, then exiting\n\n" );
    else
        fprintf( stderr, "\ncapturing cells until host memory is full, then stopping and writing to disk\n\n" );

    *want_facility_loopback = TRUE;

    return 0;
}

// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par



int flows_cleanup( void )
{
    int result;
    Bit32 count;
    BufferBlock *block, *next;

    if( !want_disk_io_during_run )
    {
        // take all the BufferBlocks off our list and save them to disk
        //
        for( block=head; block; block=next )
        {
            next = block->next;

            result = all_buffers->dump_one( block );
            if( result )
            {
                fprintf( stderr, "error %d from BufferQueue::dump_one()\n", result );
                return result;
            }

            block_count++;
#if BUFFER_DUMP_TICKS
            fprintf( stderr, "%4d\b\b\b\b", block_count );
#endif

            all_buffers->free( block );
        }
        head = NULL;
        tail = NULL;
    }

    if( file_is_open )
    {
        result = all_buffers->dump_close();
        if( result )
            fprintf( stderr, "error %d from BufferQueue::dump_close()\n", result );

        file_is_open = FALSE;
    }

    return result;
}



int flows_is_query_wanted( void )
{
    return FALSE;
}


int flows_is_query_pending( void )
{
    return FALSE;
}



// I know the extra_msg paramater isn't used,
// and I don't want the compiler to nag me about it
//
#pragma warn -par
int flows_query( const char *extra_msg )
{
    return 0;
}
#pragma warn .par



void flows_send_active( void )
{
}



// turn off compiler warning about unused parameters
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
#pragma warn -par

// this code doesn't have anything special to do
// when keys are pressed
//
void flows_keystroke( int keystroke )
{
}
// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par



void sock_init( void )
{
}



