    page 66,132

    .386p
_TEXT segment dword public use32 'CODE'
    assume cs:_TEXT
    locals @@






; differences between PCI32.ASM (this file) and PCI.ASM
;
;     "proc near" instead of "proc far" so near return
;     is generated
;
;     arguments on stack are at least 32 bits wide,
;     so offsets are +8 +12 and +16, not +6 +8 and +10
;
;     stack pointer is 32 bits wide, so we must use EBP,
;     not BP register, as the stack frame pointer
;
;     32 bit return values go in EAX, not DX:AX, so don't
;     need to do as much register manipulations to split
;     up a 32 bit result from BIOS, but do need to preserve
;     DX where we didn't before
;
;     pushing/popping less than 32 bits onto/offof stack
;     requires a prefix and is slower, so push/pop the
;     32 bit registers even if only 16 bits need to be
;     preserved
;
;     reading/writing 16 bits from stack requires a prefix
;     and is slower, and BIOS will ignore the upper bits
;     of less than 32 bit argument registers, and the stack
;     has already widened each argument to 32 bits, so just
;     read the whole 32-bit register from stack
;




; general notes:
;
;     arguments are pushed from right to left
;     when the C calling convention is used
;






; int pci_is_bios_present( void )
;
; returns 1 in AX if PCI BIOS is present
; returns 0 in AX otherwise
;
_pci_is_bios_present proc near

    ; save the caller's registers
    ;
    push ebx
    push ecx
    push edx

    ; AH = PCI function ID
    ; AL = PCI BIOS presence query
    ;
    mov ax, 0B101h
    int 1Ah

    ; should now have "PCI " in EDX register
    ;
    cmp edx, " ICP"             ; byte swapped for little-endian
    xor eax, eax                ; clear high 24 bits of eax
    setz al                     ; AL = 1 IFF zero flag set

    ; restore the caller's registers
    ;
    pop edx
    pop ecx
    pop ebx

    ret

_pci_is_bios_present endp






; unsigned short pci_get_device_handle(
;     unsigned short device_id,     // ebp + 8
;     unsigned short vendor_id,     // ebp + 12
;     unsigned short index          // ebp + 16
; )
;
; returns the handle of the Nth device with the
; desired device ID and vendor ID
;
; call this function multiple times, incrementing
; the index from zero onward, until it returns
; 0xFFFF indicating no (more) cards matched
;
; note: for Trancell Systems PCI-155 ATM network
; interface card, the device ID is 11D3h and the
; vendor ID is 0002h
;
_pci_get_device_handle proc near

    ; set up the stack frame
    ;
    push ebp
    mov ebp, esp

    ; save the caller's registers
    ;
    push ebx
    push ecx
    push edx
    push esi

    ; AH = PCI function ID
    ; AL = find device
    ;
    mov eax, 0B102h
    mov ecx, [ebp + 8]
    mov edx, [ebp + 12]
    mov esi, [ebp + 16]
    int 1Ah

    jnc handle_ok
    mov ax, 0FFFFh
    jmp handle_done

handle_ok:
    ; move the handle into AX for return to C
    ;
    ; the minimal doc I have says that device handle
    ; is just the upper 5 bits (7:3) of BL, but
    ; disassembly of TESTCM.EXE provided by Trancell
    ; uses entire contents of BX to pass to its next
    ; call to PCI BIOS, so I assume that is where
    ; the bits should stay
    ;
    movzx ax, bl

handle_done:
    ; restore the caller's registers
    ;
    pop esi
    pop edx
    pop ecx
    pop ebx

    ; clean up the stack frame
    ;
    pop ebp

    ; caller cleans up the stack themselves
    ;
    ret

_pci_get_device_handle endp






; Bit32 pci_read_config_32(
;     Bit16 device_handle,     // ebp + 8
;     Bit16 offset             // ebp + 12
; )
;
; returns the 32-bit configuration register
; on the PCI device with the given handle
; at the given offset
;
; errors are ignored because PCI/host bridge
; is supposed to return 0xFFFF for vendor ID
; for non-existent devices (master abort will
; be seen, but is not considered an error)
;
_pci_read_config_32 proc near

    ; set up the stack frame
    ;
    push ebp
    mov ebp, esp

    ; save the caller's registers
    ;
    push ebx
    push ecx
    push edi

    ; AH = PCI function ID
    ; AL = read configuration DWORD
    ; BX 15:8 = bus number
    ;     7:3 = device ID
    ;     2:0 = function number (subdevice)
    ; DI = byte offset into config space
    ;
    mov eax, 0B10Ah
    mov ebx, [ebp + 8]
    mov edi, [ebp + 12]
    int 1Ah

if 1
    ; take care of errors (carry set) by changing the
    ; register we will be returning to be negative 1
    ;
    jnc @@1
    mov ecx, 0FFFFFFFFh

@@1:
endif

    ; move the value into EAX for return to 32-bit C language
    ;
    mov eax, ecx

    ; restore the caller's registers
    ;
    pop edi
    pop ecx
    pop ebx

    ; clean up the stack frame
    ;
    pop ebp

    ; caller cleans up the stack themselves
    ;
    ret

_pci_read_config_32 endp








; Bit16 pci_read_config_16(
;     Bit16 device_handle,     // ebp + 8
;     Bit16 offset             // ebp + 12
; )
;
; returns the 16-bit configuration register
; on the PCI device with the given handle
; at the given offset
;
; errors are ignored
;
_pci_read_config_16 proc near

    ; set up the stack frame
    ;
    push ebp
    mov ebp, esp

    ; save the caller's registers
    ;
    push ebx
    push ecx
    push edi

    ; AH = PCI function ID
    ; AL = read configuration WORD
    ; BX 15:8 = bus number
    ;     7:3 = device ID
    ;     2:0 = function number (subdevice)
    ; DI = byte offset into config space
    ;
    mov eax, 0B109h
    mov ebx, [ebp + 8]
    mov edi, [ebp + 12]
    int 1Ah

if 1
    ; take care of errors (carry set) by changing the
    ; register we will be returning to be negative 1
    ;
    jnc @@1
    mov cx, 0FFFFh

@@1:
endif

    ; move the value into AX for return to C language
    ;
    mov ax, cx

    ; restore the caller's registers
    ;
    pop edi
    pop ecx
    pop ebx

    ; clean up the stack frame
    ;
    pop ebp

    ; caller cleans up the stack themselves
    ;
    ret

_pci_read_config_16 endp








; Bit8 pci_read_config_8(
;     Bit16 device_handle,     // ebp + 8
;     Bit16 offset             // ebp + 12
; )
;
; returns the 8-bit configuration register
; on the PCI device with the given handle
; at the given offset
;
; errors are ignored
;
_pci_read_config_8 proc near

    ; set up the stack frame
    ;
    push ebp
    mov ebp, esp

    ; save the caller's registers
    ;
    push ebx
    push ecx
    push edi

    ; AH = PCI function ID
    ; AL = read configuration BYTE
    ; BX 15:8 = bus number
    ;     7:3 = device ID
    ;     2:0 = function number (subdevice)
    ; DI = byte offset into config space
    ;
    mov eax, 0B108h
    mov ebx, [ebp + 8]
    mov edi, [ebp + 12]
    int 1Ah

if 1
    ; take care of errors (carry set) by changing the
    ; register we will be returning to be negative 1
    ;
    jnc @@1
    mov cl, 0FFh

@@1:
endif

    ; move the value into AL for return to C language
    ;
    mov al, cl

    ; restore the caller's registers
    ;
    pop edi
    pop ecx
    pop ebx

    ; clean up the stack frame
    ;
    pop ebp

    ; caller cleans up the stack themselves
    ;
    ret

_pci_read_config_8 endp








; void pci_write_config_32(
;     Bit16 device_handle,     // ebp + 8
;     Bit16 offset,            // ebp + 12
;     Bit32 value              // ebp + 16
; )
;
; sets the 32-bit configuration register
; on the PCI device with the given handle
; at the given offset to the given value
;
; errors are ignored
;
_pci_write_config_32 proc near

    ; set up the stack frame
    ;
    push ebp
    mov ebp, esp

    ; save the caller's registers
    ;
    push eax
    push ebx
    push ecx
    push edi

    ; AH = PCI function ID
    ; AL = write configuration DWORD
    ; BX 15:8 = bus number
    ;     7:3 = device ID
    ;     2:0 = function number (subdevice)
    ; DI = byte offset into config space
    ; ECX = value to write
    ;
    mov ax, 0B10Dh
    mov bx, [ebp + 8]
    mov di, [ebp + 12]
    mov ecx, [ebp + 16]
    int 1Ah

    ; restore the caller's registers
    ;
    pop edi
    pop ecx
    pop ebx
    pop eax

    ; clean up the stack frame
    ;
    pop ebp

    ; caller cleans up the stack themselves
    ;
    ret

_pci_write_config_32 endp








; void pci_write_config_16(
;     Bit16 device_handle,     // ebp + 8
;     Bit16 offset,            // ebp + 12
;     Bit16 value              // ebp + 16
; )
;
; sets the 16-bit configuration register
; on the PCI device with the given handle
; at the given offset to the given value
;
; errors are ignored
;
_pci_write_config_16 proc near

    ; set up the stack frame
    ;
    push ebp
    mov ebp, esp

    ; save the caller's registers
    ;
    push eax
    push ebx
    push ecx
    push edi

    ; AH = PCI function ID
    ; AL = write configuration WORD
    ; BX 15:8 = bus number
    ;     7:3 = device ID
    ;     2:0 = function number (subdevice)
    ; DI = byte offset into config space
    ; CX = value to write
    ;
    mov eax, 0B10Ch
    mov ebx, [ebp + 8]
    mov edi, [ebp + 12]
    mov ecx, [ebp + 16]
    int 1Ah

    ; restore the caller's registers
    ;
    pop edi
    pop ecx
    pop ebx
    pop eax

    ; clean up the stack frame
    ;
    pop ebp

    ; caller cleans up the stack themselves
    ;
    ret

_pci_write_config_16 endp








; void pci_write_config_8(
;     Bit16 device_handle,     // ebp + 8
;     Bit16 offset,            // ebp + 12
;     Bit8 value               // ebp + 16
; )
;
; sets the 8-bit configuration register
; on the PCI device with the given handle
; at the given offset to the given value
;
; errors are ignored
;
_pci_write_config_8 proc near

    ; set up the stack frame
    ;
    push ebp
    mov ebp, esp

    ; save the caller's registers
    ;
    push eax
    push ebx
    push ecx
    push edi

    ; AH = PCI function ID
    ; AL = write configuration BYTE
    ; BX 15:8 = bus number
    ;     7:3 = device ID
    ;     2:0 = function number (subdevice)
    ; DI = byte offset into config space
    ; CL = value to write
    ;
    mov eax, 0B10Bh
    mov ebx, [ebp + 8]
    mov edi, [ebp + 12]
    mov ecx, [ebp + 16]
    int 1Ah

    ; restore the caller's registers
    ;
    pop edi
    pop ecx
    pop ebx
    pop eax

    ; clean up the stack frame
    ;
    pop ebp

    ; caller cleans up the stack themselves
    ;
    ret

_pci_write_config_8 endp








public _pci_is_bios_present
public _pci_get_device_handle
public _pci_read_config_32
public _pci_read_config_16
public _pci_read_config_8
public _pci_write_config_32
public _pci_write_config_16
public _pci_write_config_8






_TEXT ends

    end
