#include <stdio.h>                      // printf()
#include <stdlib.h>                     // exit()
#include "..\pci\pci.h"



// find the FORE Systems ATM NIC's on the PCI bus
// and initialize them
//
void main( void )
{
    unsigned count;
    Bit8 value;
    Bit16 handle;
    Bit32 address;
    unsigned irq_level;
    int result;
    Bit8 revision;
    Bit16 bus, device, function;
    Bit32 device_and_vendor;

    // check whether PCI BIOS is present
    //
    if( !pci_is_bios_present() )
    {
        printf( "Sorry, PCI BIOS is not present.\n" );
        exit( 1 );
    }

    printf( "searching for all valid PCI devices\n" );

    // for all PCI busses...
    //
    for( bus=0; bus<PCI_BUSSES_PER_MACHINE; bus++ )
    {
        // for all PCI devices on this bus...
        //
        for( device=0; device<PCI_DEVICES_PER_BUS; device++ )
        {
            // for all functions on this device...
            //
            for( function=0; function<PCI_FUNCTIONS_PER_DEVICE; function++ )
            {
                // calculate the handle from bus, device, and function
                //
                handle = (bus << 8) | (device << 3) | function;

                // get the device ID and vendor ID together,
                // and also check if this is a valid device
                //
                device_and_vendor = pci_read_config_32( handle, 0 );

                if( device_and_vendor == (Bit32)-1 )
                    continue;

                revision = pci_get_revision( handle );
                address = pci_get_base_address( handle );
                irq_level = pci_get_irq_level( handle );

                printf( "bus=%d device=%d func=%d (handle=0x%4.4X)"
                    " saw device ID 0x%4.4X vendor ID 0x%4.4X revision 0x%.2X"
                    " phys address 0x%8.8lX IRQ level %d\n",
                    bus, device, function, handle,
                    device_and_vendor >> 16,
                    device_and_vendor & 65535,
                    revision, address, irq_level
                );
            }
        }
    }

    printf( "PCI configuration discovery done\n\n" );
}



