#include <stdio.h>                      // printf()
#include <mem.h>                        // memset()
#include <stdlib.h>                     // strtoul()
#include <conio.h>                      // kbhit()
#include "../scsi/aspi.h"               // AspiSrb aspi
#include "../intel/byteswap.h"          // htonl()
#include "../intel/timestmp.h"          // PentiumClock
#include "../dpmi32/realmode.h"         // real_to_physical()



#define MAX_IO          10
#define MAX_BLOCKS      256
#define BLOCK_SIZE      512
#define TIMEOUT         100.0



// don't bother to allocate all the memory someone
// might want to write, since we cannot touch memory
// above the 1 MB line in realmode, and the max for
// a single SCSI command is 65535 blocks, so even if
// we leave the disk formatted for 512 byte blocks,
// that is 32 MB minus 512 bytes
//
// it is safe to refer to more than we allocated
// because we are telling ASPI to write, not read
//
Bit8 huge my_data0[ MAX_BLOCKS ][ BLOCK_SIZE ];
Bit32 my_data = real_to_physical( (Bit32) my_data0 );

Bit32 iterations[ MAX_IO ];
PentiumClock start_times[ MAX_IO ];
PentiumClock   end_times[ MAX_IO ];
AspiExecuteScsiIo io_srbs[ MAX_IO ];
Bit32 blocks;
Bit32 iterations_wanted;
unsigned volatile posting_call_count = 0;




// called when one of our SCSI I/O SRB's completes
//
// using printf() or in general any nonessential operation
// (something that doesn't absolutely have to happen when
// the I/O completes and before everybody else's does)
// detracts from our timing of the I/O operation, so DON'T!
//
// using floating point instructions here causes a lockup!
// so cannot convert elapsed time to seconds, or divide it into bytes
//
void far __loadds __saveregs notifier( AspiExecuteScsiIo far *my_scsi_io )
{
    unsigned count = my_scsi_io - io_srbs;

    // if we haven't hit our iteration limit...
    //
    if( ++iterations[ count ] < iterations_wanted )
    {
        // if the command that just completed had no error...
        //
        if( my_scsi_io->Status == AspiStatusScsiRequestCompletedWithoutError )
        {
            // start another write to where we left off on the same disk,
            // but rewinding to the beginning of the same host buffer
            // (which incidentally all the disks are sharing)
            //
            // we only touch those fields that will have been altered
            // by a successful execution of the previous write
            //
            // we don't mess with most fields of CDB since it wasn't
            // overwritten by sense data (error details)
            //
            // assumes byte residue not requested, so doesn't refill
            // SRB length
            //
            ScsiWriteExtended *my_write = &my_scsi_io->CdbAndSense.Cdb.WriteExtended;
            my_write->LogicalBlockAddress = htonl( blocks + ntohl( my_write->LogicalBlockAddress ) );
            my_scsi_io->Status = AspiStatusScsiRequestInProgress;
            aspi( (AspiSrb *)my_scsi_io );
            return;
        }
    }

    posting_call_count++;
    end_times[ count ].set();
}



int main( int argc, char *argv[] )
{
    AspiSrb my_srb;
    AspiExecuteScsiIo *my_scsi_io;
    ScsiWriteExtended *my_write;
    unsigned count;
    const char *program = argv[ 0 ];
    const char *offset_string = argv[ 1 ];
    const char *blocks_string = argv[ 2 ];
    const char *iterations_wanted_string = argv[ 3 ];
    const char *max_per_bus_string = argv[ 4 ];
    Bit32 offset;
    PentiumClock now;
    unsigned latest;
    unsigned max_per_bus;

    if( argc < 4 )
    {
        fprintf( stderr, "usage: %s block_offset block_count iterations {max_per_bus}\n", program );
        return 1;
    }
    blocks = strtoul( blocks_string, NULL, 10 );
    if( blocks > BIT16_MAX )
    {
        printf( "sorry, %lu is bigger than maximum block count of %u\n"
            "you might want to consider using 0 which means 'until end of disk'\n",
            blocks, BIT16_MAX
        );
        return 2;
    }
    offset = strtoul( offset_string, NULL, 10 );
    iterations_wanted = strtoul( iterations_wanted_string, NULL, 10 );
    if( argc > 4 )
        max_per_bus = atoi( max_per_bus_string );
    else
        max_per_bus = MAX_SCSI_DEVICES_PER_BUS;
    printf( "writing data to %u blocks %lu times starting at offset %lu"
        " on up to %d disks per SCSI bus\n",
        blocks, iterations_wanted, offset, max_per_bus );

    if( aspi_find_disks( &my_srb, max_per_bus ) )
        return 3;

    // don't bother trying to set all the buffer to a
    // known value, since there isn't a realmode routine
    // to touch all of it, only 64KB at a time, only
    // below the 1 MB line
    //
    //memset( &my_data, 0x67, MAX_BLOCKS*BLOCK_SIZE );

    // fill in the data structures for all I/O requests
    //
    for( count=0; count<aspi_disk_count; count++ )
    {
        my_scsi_io = io_srbs + count;
        my_write = &my_scsi_io->CdbAndSense.Cdb.WriteExtended;

        printf( "%dth SRB starting write to disk at SCSI bus/ID/LUN %d/%d/%d\n",
            count + 1,
            aspi_all_disks[count].HostAdaptorNumber,
            aspi_all_disks[count].TargetId,
            aspi_all_disks[count].Lun
        );

        // try writing same data to many logical blocks
        //
        memset( my_scsi_io, 0, sizeof(AspiExecuteScsiIo) );
        my_scsi_io->Command = AspiCommandExecuteScsiIo;
        my_scsi_io->HostAdaptorNumber = aspi_all_disks[count].HostAdaptorNumber;
        my_scsi_io->RequestFlags.EnablePosting = TRUE;
        my_scsi_io->RequestFlags.EnableResidualByteLengthReporting = FALSE;
        my_scsi_io->RequestFlags.TransferDirection = AspiXferDirectionHostToScsiTarget;
        my_scsi_io->RequestFlags.Reserved = 0;
        my_scsi_io->TargetId = aspi_all_disks[count].TargetId;;
        my_scsi_io->Lun = aspi_all_disks[count].Lun;
        my_scsi_io->DataAllocationLength = blocks*BLOCK_SIZE;
        my_scsi_io->SenseAllocationLength = sizeof( my_scsi_io->CdbAndSense.Sense );
        my_scsi_io->DataBufferPointer = my_data + count*blocks*BLOCK_SIZE;
        my_scsi_io->SrbLinkPointer = 0;         // we won't be linking
        my_scsi_io->PostRoutineAddress = (Bit32) notifier;
        my_scsi_io->CdbLength = sizeof( *my_write );
        my_write->Command = ScsiCommandWriteExtended;
        my_write->LogicalBlockAddress = htonl( offset );
        my_write->TransferLength = htons( blocks );

        iterations[ count ] = 0;
    }

    // now start up all the I/O's at once
    //
    for( count=0; count<aspi_disk_count; count++ )
    {
        start_times[count].set();

        aspi( (AspiSrb *) (io_srbs + count) );
    }

    // wait for posting routine to be called for all disks
    //
    // (don't print any message here, to save CPU time)
    // (don't call kbhit() to save CPU time)
    //
    //printf( "waiting for keypress or all %u SRB's to be posted\n", aspi_disk_count );
    while( posting_call_count < aspi_disk_count )
#if 0
        if( kbhit() )
            break;
#else
        ;
#endif

    latest = 0;
    for( count=0; count<aspi_disk_count; count++ )
    {
        // remember which SRB ended later
        //
        if( end_times[ latest ] < end_times[ count ] )
            latest = count;

        my_scsi_io = io_srbs + count;
        if( my_scsi_io->Status != AspiStatusScsiRequestCompletedWithoutError )
        {
            printf( "%dth SRB iteration %u error: status was %d\n", count+1, iterations[count], my_scsi_io->Status );
        }
        else
        {
            printf( "%dth SRB residual length was %lu bytes\n", count+1, my_scsi_io->DataAllocationLength );

            now = end_times[ count ] - start_times[ count ];
            printf( "%dth SRB %g million bytes per second\n", count+1,
                iterations[count] * BLOCK_SIZE * blocks / now.get_seconds() / 1e6 );
        }
    }

    printf( "%uth SRB actually completed last\n", latest+1 );

    now = end_times[ latest ] - start_times[ 0 ];
    printf( "assuming all iterations were successful,\n"
        " aggregate disk rate was %g million bytes per second\n",
        aspi_disk_count * iterations_wanted * BLOCK_SIZE * blocks / now.get_seconds() / 1e6
    );

    printf( "all done!\n" );

    return 0;
}
