#include <stdio.h>                      // printf()
#include <mem.h>                        // memset()
#include <stdlib.h>                     // strtoul()
#include <conio.h>                      // kbhit()
#include <string.h>                     // stricmp()
#include "../scsi/aspi.h"               // AspiSrb aspi
#include "../intel/byteswap.h"          // htonl()
#include "../intel/timestmp.h"          // PentiumClock
#include "../dpmi32/realmode.h"         // real_to_physical()



#define TIMEOUT         1.0



struct
{
    ScsiMode10ParmList Parms;
    ScsiModeCaching Caching;
} my_data;

PentiumClock end_time;

int volatile posting_occurred;

typedef enum
{
    WantReadCacheDisable = 0,
    WantMultiplicationFactor,
    WantWriteCacheEnable,
    WantPrefetchAcrossTimeDiscontinuities,
    WantCachingAnalysisPermitted,
    WantAbortPrefetchUponSelection,
    WantInitiatorControl,
    WantDemandReadRetentionPriority,
    WantWriteRetentionPriority,
    WantDisablePrefetchTransferLength,
    WantMinimumPrefetch,
    WantMaximumPrefetch,
    WantMaximumPrefetchCeiling,
    WantDisableReadAhead,
    WantLogicalBlockCacheSegmentSize,
    WantForceSequentialWrite,
    WantNumberOfCacheSegments,
    WantCacheSegmentSize,
    WantNonCacheSegmentSize,
    WantAllParms
} WhichParmWanted;

const char *all_parm_names[] =
{
    "ReadCacheDisable",
    "MultiplicationFactor",
    "WriteCacheEnable",
    "PrefetchAcrossTimeDiscontinuities",
    "CachingAnalysisPermitted",
    "AbortPrefetchUponSelection",
    "InitiatorControl",
    "DemandReadRetentionPriority",
    "WriteRetentionPriority",
    "DisablePrefetchTransferLength",
    "MinimumPrefetch",
    "MaximumPrefetch",
    "MaximumPrefetchCeiling",
    "DisableReadAhead",
    "LogicalBlockCacheSegmentSize",
    "ForceSequentialWrite",
    "NumberOfCacheSegments",
    "CacheSegmentSize",
    "NonCacheSegmentSize",
    "AllParms",
    ""
};




// I know the following function doesn't use its parameter
// and I don't want the compiler to bitch about it
//
#pragma warn -par

// called when one of our SCSI I/O SRB's completes
//
// using printf() or in general any nonessential operation
// (something that doesn't absolutely have to happen when
// the I/O completes and before everybody else's does)
// detracts from our timing of the I/O operation, so DON'T!
//
// using floating point instructions here causes a lockup!
// so cannot convert elapsed time to seconds, or divide it into bytes
//
void far __loadds __saveregs notifier( AspiExecuteScsiIo far *my_scsi_io )
{
    posting_occurred = TRUE;
    end_time.set();
}

// enable compiler warning about unused function parms
// if it used to be on
//
#pragma warn .par



int usage( const char *program )
{
    unsigned count;

    fprintf( stderr, "usage: %s parm_name {parm_value {save_to_disk}}\n", program );
    fprintf( stderr, "    where (optional) parm_value is an integer\n" );
    fprintf( stderr, "    and save_to_disk is any text (otherwise only current values affected)\n" );
    fprintf( stderr, "    and (case-insensitive) parm_name is one of:\n" );
    for( count=0; all_parm_names[count][0]!='\0'; count++ )
        fprintf( stderr, "\t%s\n", all_parm_names[count] );

    return 1;
}



int main( int argc, char *argv[] )
{
    AspiSrb my_srb;
    AspiExecuteScsiIo *my_scsi_io = &my_srb.ExecuteScsiIo;
    ScsiModeSense10 *my_sense = &my_scsi_io->CdbAndSense.Cdb.ModeSense10;
    ScsiModeSelect10 *my_select = &my_scsi_io->CdbAndSense.Cdb.ModeSelect10;
    unsigned count;
    const char *program = argv[ 0 ];
    const char *parm = argv[ 1 ];
    const char *value_string = argv[ 2 ];
    int found_parm;
    Bit32 value, old_value;
    int want_value_changed = FALSE;
    int want_saved_to_disk = FALSE;
    WhichParmWanted parm_wanted;
    WhichParmWanted parm_index;
    WhichParmWanted parm_min, parm_max;
    PentiumClock start_time, now;
    unsigned latest;

    if( argc < 2 )
        return usage( program );

    // search for the parm name in the list of valid ones
    //
    found_parm = FALSE;
    for( parm_wanted=(WhichParmWanted)0; all_parm_names[parm_wanted][0]!='\0'; parm_wanted++ )
    {
        if( !stricmp( parm, all_parm_names[parm_wanted] ) )      // equal?
        {
            found_parm = TRUE;
            break;
        }
    }
    if( !found_parm )
    {
        fprintf( stderr, "sorry, '%s' is not a valid parameter name\n", parm );
        usage( program );
        return 2;
    }

    // parse parm value
    //
    if( argc > 2 )
    {
        want_value_changed = TRUE;
        value = strtoul( value_string, NULL, 10 );
        printf( "selecting '%s' to have current value %lu\n", parm, value );
        if( argc > 3 )
        {
            want_saved_to_disk = TRUE;
            printf( "also saving to disk as powerup value\n" );
        }
    }
    else
    {
        printf( "just sensing old value of '%s'\n", parm );
    }

    if( aspi_find_disks( &my_srb, MAX_SCSI_DEVICES_PER_BUS, MAX_SCSI_BUSES ) )
        return 3;

    // for all drives...
    //
    for( count=0; count<aspi_disk_count; count++ )
    {
        printf( "%dth SRB looking at disk at SCSI bus/ID/LUN %d/%d/%d\n",
            count + 1,
            aspi_all_disks[count].HostAdaptorNumber,
            aspi_all_disks[count].TargetId,
            aspi_all_disks[count].Lun
        );

        // get current data from drive
        //
        // assumes ASPI driver has been patched to accept physical memory
        // addresses instead of realmode segment:offset addresses
        //
        memset( my_scsi_io, 0, sizeof(AspiExecuteScsiIo) );
        my_scsi_io->Command = AspiCommandExecuteScsiIo;
        my_scsi_io->HostAdaptorNumber = aspi_all_disks[count].HostAdaptorNumber;
        my_scsi_io->RequestFlags.EnablePosting = TRUE;
        my_scsi_io->RequestFlags.EnableResidualByteLengthReporting = FALSE;
        my_scsi_io->RequestFlags.TransferDirection = AspiXferDirectionScsiTargetToHost;
        my_scsi_io->RequestFlags.Reserved = 0;
        my_scsi_io->TargetId = aspi_all_disks[count].TargetId;;
        my_scsi_io->Lun = aspi_all_disks[count].Lun;
        my_scsi_io->DataAllocationLength = sizeof( my_data );
        my_scsi_io->SenseAllocationLength = sizeof( my_scsi_io->CdbAndSense.Sense );
        my_scsi_io->DataBufferPointer = real_to_physical( (Bit32) &my_data );
        my_scsi_io->SrbLinkPointer = 0;         // we won't be linking
        my_scsi_io->PostRoutineAddress = (Bit32) notifier;
        my_scsi_io->CdbLength = sizeof( *my_sense );
        my_sense->Command = ScsiCommandModeSense10;
        my_sense->DisableBlockDescriptors = TRUE;
        my_sense->PageCode = ScsiModePageCaching;
        my_sense->PageControl = ScsiModePageControlCurrent;
        my_sense->AllocationLength = htons( sizeof( my_data ) );
        posting_occurred = FALSE;
        start_time.set();
        aspi( &my_srb );

        while( !posting_occurred )
            if( kbhit() )
            {
                getch();
                printf( "mode sense (10) aborted by keypress\n" );
                return 4;
            }

        end_time -= start_time;
        printf( "%g milliseconds elapsed   sensing current caching parms\n", 1000*end_time.get_seconds() );

        // validate results of mode sense
        //
        if( my_scsi_io->Status != AspiStatusScsiRequestCompletedWithoutError )
        {
            printf( "ASPI returned unsuccessful status %d\n", my_scsi_io->Status );
            printf( "host adaptor status was %d\n", my_scsi_io->HostAdaptorStatus );
            printf( "target status was %d\n", my_scsi_io->TargetStatus );
            return 5;
        }
        if( ntohs( my_data.Parms.SenseDataLength ) != sizeof( my_data ) - sizeof( my_data.Parms.SenseDataLength ) )
        {
            printf( "internal error: drive returned %u bytes of sense data,"
                " not the %u expected for just parmlist and caching page, minus %d byte length itself\n",
                ntohs( my_data.Parms.SenseDataLength ),
                sizeof( my_data ) - sizeof( my_data.Parms.SenseDataLength ),
                sizeof( my_data.Parms.SenseDataLength )
            );
            return 6;
        }
        if( my_data.Parms.BlockDescriptorLength != 0 )
        {
            printf( "internal error: drive returned %u bytes of blocks"
                " descriptor data, not the 0 expected\n",
                ntohs( my_data.Parms.BlockDescriptorLength )
            );
            return 7;
        }
        if( my_data.Caching.PageCode != ScsiModePageCaching )
        {
            printf( "internal error: drive returned page %d, not page %d (caching info)\n",
                my_data.Caching.PageCode,
                ScsiModePageCaching
            );
            return 8;
        }
        if( my_data.Caching.PageLength != sizeof( ScsiModeCaching ) - 2 )
        {
            printf( "internal error: drive returned page length %d, not %d expected for just caching info minus 2 byte header\n",
                my_data.Caching.PageLength,
                sizeof( ScsiModeCaching ) - 2
            );
            return 9;
        }

        // dump current parm of interest
        //
        if( parm_wanted == WantAllParms )
        {
            parm_min = (WhichParmWanted)0;
            parm_max = WantAllParms;
            if( want_value_changed )
                printf( "sorry, old values cannot be changed when you wildcard\n" );
            want_value_changed = FALSE;
        }
        else
        {
            parm_min = parm_wanted;
            parm_max = (WhichParmWanted) (parm_wanted + 1);
        }

        for( parm_index=parm_min; parm_index<parm_max; parm_index++ )
        {
            switch( parm_index )
            {
            case WantReadCacheDisable:
                old_value = my_data.Caching.ReadCacheDisable;
                if( want_value_changed )
                    my_data.Caching.ReadCacheDisable = value;
                break;

            case WantMultiplicationFactor:
                old_value = my_data.Caching.MultiplicationFactor;
                if( want_value_changed )
                    my_data.Caching.MultiplicationFactor = value;
                break;

            case WantWriteCacheEnable:
                old_value = my_data.Caching.WriteCacheEnable;
                if( want_value_changed )
                    my_data.Caching.WriteCacheEnable = value;
                break;

            case WantPrefetchAcrossTimeDiscontinuities:
                old_value = my_data.Caching.PrefetchAcrossTimeDiscontinuities;
                if( want_value_changed )
                    my_data.Caching.PrefetchAcrossTimeDiscontinuities = value;
                break;

            case WantCachingAnalysisPermitted:
                old_value = my_data.Caching.CachingAnalysisPermitted;
                if( want_value_changed )
                    my_data.Caching.CachingAnalysisPermitted = value;
                break;

            case WantAbortPrefetchUponSelection:
                old_value = my_data.Caching.AbortPrefetchUponSelection;
                if( want_value_changed )
                    my_data.Caching.AbortPrefetchUponSelection = value;
                break;

            case WantInitiatorControl:
                old_value = my_data.Caching.InitiatorControl;
                if( want_value_changed )
                    my_data.Caching.InitiatorControl = value;
                break;

            case WantDemandReadRetentionPriority:
                old_value = my_data.Caching.DemandReadRetentionPriority;
                if( want_value_changed )
                    my_data.Caching.DemandReadRetentionPriority = value;
                break;

            case WantWriteRetentionPriority:
                old_value = my_data.Caching.WriteRetentionPriority;
                if( want_value_changed )
                    my_data.Caching.WriteRetentionPriority = value;
                break;

            case WantDisablePrefetchTransferLength:
                old_value = ntohs( my_data.Caching.DisablePrefetchTransferLength );
                if( want_value_changed )
                    my_data.Caching.DisablePrefetchTransferLength = htons( value );
                break;

            case WantMinimumPrefetch:
                old_value = ntohs( my_data.Caching.MinimumPrefetch );
                if( want_value_changed )
                    my_data.Caching.MinimumPrefetch = htons( value );
                break;

            case WantMaximumPrefetch:
                old_value = ntohs( my_data.Caching.MaximumPrefetch );
                if( want_value_changed )
                    my_data.Caching.MaximumPrefetch = htons( value );
                break;

            case WantMaximumPrefetchCeiling:
                old_value = ntohs( my_data.Caching.MaximumPrefetchCeiling );
                if( want_value_changed )
                    my_data.Caching.MaximumPrefetchCeiling = htons( value );
                break;

            case WantDisableReadAhead:
                old_value = my_data.Caching.DisableReadAhead;
                if( want_value_changed )
                    my_data.Caching.DisableReadAhead = value;
                break;

            case WantLogicalBlockCacheSegmentSize:
                old_value = my_data.Caching.LogicalBlockCacheSegmentSize;
                if( want_value_changed )
                    my_data.Caching.LogicalBlockCacheSegmentSize = value;
                break;

            case WantForceSequentialWrite:
                old_value = my_data.Caching.ForceSequentialWrite;
                if( want_value_changed )
                    my_data.Caching.ForceSequentialWrite = value;
                break;

            case WantNumberOfCacheSegments:
                old_value = my_data.Caching.NumberOfCacheSegments;
                if( want_value_changed )
                    my_data.Caching.NumberOfCacheSegments = value;
                break;

            case WantCacheSegmentSize:
                old_value = ntohs( my_data.Caching.CacheSegmentSize );
                if( want_value_changed )
                {
                    my_data.Caching.SizeEnable = TRUE;
                    my_data.Caching.CacheSegmentSize = htons( value );
                }
                break;

            case WantNonCacheSegmentSize:
                old_value = my_data.Caching.NonCacheSegmentSize[0] << 16
                          | my_data.Caching.NonCacheSegmentSize[1] << 8
                          | my_data.Caching.NonCacheSegmentSize[2];
                if( want_value_changed )
                {
                    my_data.Caching.NonCacheSegmentSize[0] = value >> 16;
                    my_data.Caching.NonCacheSegmentSize[1] = value >> 8;
                    my_data.Caching.NonCacheSegmentSize[2] = value;
                }
                break;
            };

            printf( "old value  %5lu for '%s'\n", old_value, all_parm_names[parm_index] );
        }

        if( !want_value_changed )
            continue;

        // put the current (and maybe the powerup) values to drive
        //
        // assumes ASPI driver has been patched to accept physical memory
        // addresses instead of realmode segment:offset addresses
        //
        my_scsi_io->Status = AspiCommandExecuteScsiIo;
        my_scsi_io->Status = AspiStatusScsiRequestInProgress;
        my_scsi_io->CdbLength = sizeof( *my_select );
        memset( my_select, 0, sizeof(*my_select) );
        my_select->Command = ScsiCommandModeSelect10;
        my_select->SaveModeParameters = want_saved_to_disk;
        my_select->PageFormat = TRUE;
        my_select->ParmListLength = htons( sizeof( my_data ) );
        posting_occurred = FALSE;
        start_time.set();
        aspi( &my_srb );

        while( !posting_occurred )
            if( kbhit() )
            {
                getch();
                printf( "mode select (10) aborted by keypress\n" );
                return 10;
            }

        end_time -= start_time;
        printf( "%g milliseconds elapsed selecting current caching parms\n", 1000*end_time.get_seconds() );

        // validate results of mode select
        //
        if( my_scsi_io->Status != AspiStatusScsiRequestCompletedWithoutError )
        {
            printf( "ASPI returned unsuccessful status %d\n", my_scsi_io->Status );
            printf( "host adaptor status was %d\n", my_scsi_io->HostAdaptorStatus );
            printf( "target status was %d\n", my_scsi_io->TargetStatus );
            return 11;
        }

        printf( "changed to %5lu\n", value );
    }

    printf( "all done!\n" );

    return 0;
}
