/* 1955, Sun 4 Jan 98

   wattcp monitoring routines:
      output in format similar to tcpdump.

   Nevil Brownlee, The University of Auckland */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "..\intel\timestmp.h"
#include "..\wattcp\wattcp.h"
#include "..\wattcp\elib.h"
#include "..\types\types.h"
#include "..\dpmi32\phys.h"

extern void (*_dbugxmit)(
   sock_type *s, in_Header *ih, tcp_Header *ph, char *line);
extern void (*_dbugrecv)(
   sock_type *s, in_Header *ih, tcp_Header *ph, char *line);

static trace_level;

void p_upticks(void)
{
   PentiumClock now;
   double s;  Bit32 t;
   now.set();
   s = now.get_seconds();  t = now.get_ticks();
   printf("%.3f [%4u] ", s, t%10000);
   }

#define TCP_TX  1
#define TCP_RX  2

Bit32 rx_seqbase = 0;
Bit32 tx_seqbase = 0;

void p_tcp_pkt(int type,
   tcp_Socket *s, in_Header *ip, tcp_Header *tp, char *line)
{
   char *msg[] = {"-> ", " <-"};
   int len, d_offset;
   int flags = ntohs(tp->flags);
   Bit32 seq = ntohl(tp->seqnum);
   Bit32 ack = ntohl(tp->acknum);

   p_upticks();
   printf("%s ", msg[type-1]);
   if (flags & tcp_FlagSYN) printf("S");
   if (flags & tcp_FlagFIN) printf("F");
   if (flags & tcp_FlagRST) printf("R");
   if (flags & tcp_FlagPUSH) printf("P");
   if ( (flags & 
         (tcp_FlagSYN | tcp_FlagFIN | tcp_FlagRST | tcp_FlagPUSH) )
         == 0)
      printf(".");

   len = in_GetHdrlenBytes(ip);  // Bytes in IP header
   len = ntohs(ip->length) - len;  // Bytes in TCP segment
   d_offset = tcp_GetDataOffset(tp) << 2;  // Bytes in TCP header
   len -= d_offset;  // Bytes in TCP (options and) data

   if (type == TCP_TX) {
      printf(" %lu:%lu(%u)", seq-tx_seqbase, seq-tx_seqbase + len, len);
      }
   else {
      printf(" %lu:%lu(%u)", seq-rx_seqbase, seq-rx_seqbase + len, len);
      }
   if (flags & tcp_FlagACK) {
      printf(" ack %lu win %u\n", 
         type == TCP_TX ? ack-rx_seqbase : ack-tx_seqbase,
         ntohs(tp->window));
      }
   else printf("\n");

   if (flags & tcp_FlagSYN) {
      if (type == TCP_TX) tx_seqbase = seq;
      else rx_seqbase = seq;
      }
   }

static void dbugxmit(sock_type *st, in_Header *ih, tcp_Header *ph, char *line)
{
   if (st == NULL) return;
   if (st->tcp.ip_type != IPPROTO_TCP) return;
   p_tcp_pkt(TCP_TX, &st->tcp, ih, ph, line);
   }

static void dbugrecv(sock_type *st, in_Header *ih, tcp_Header *ph, char *line)
{
   if (st == NULL) return;
   if (st->tcp.ip_type != IPPROTO_TCP) return;
   p_tcp_pkt(TCP_RX, &st->tcp, ih, ph, line);
   }

void init_trace(int level)
{
   if (level == 0)
      _dbugxmit = _dbugrecv = NULL;
   else {
      _dbugxmit = dbugxmit;
      _dbugrecv = dbugrecv;
      }
   trace_level = level;
   return;
   }
