
// FRAGSUPPORT enables support for packet reassembly of fragmented packets
#define noFRAGSUPPORT

#define PKT_DEBUG     0
#define IP_BUF_TRACE  0

#pragma inline

#include "..\wattcp\copyrigh.h"
#include "..\wattcp\wattcp.h"
#include "..\wattcp\elib.h"
#include "..\wattcp\ethdev.h"
#include "..\wattcp\pktdrv.h"

#include <dos.h>
#ifdef __BORLANDC__
#include <mem.h>
#endif
#include <string.h>
#include <stdlib.h>

void hwt_init(void);  /* Initialise timer routines (in ..\intel\timestmp.cpp) */
void hwt_fin(void);  /* Declared here to avoid compiling pcpkt as a C++ module */

#pragma inline            /* 13 Jul 93 */

#define noNEVILDEBUG

#define MAXBUFS  12  /* Max nbr of Ethernet buffers  (Nevil) */

char au_monitor = 0;  /* AU: 1 to enable monitoring */

#define INT_FIRST   0x60
#define INT_LAST    0x80
#define PD_DRIVER_INFO  0x01FF
#define PD_ACCESS       0x0200
#define PD_RELEASE      0x0300
#define PD_SEND         0x0400
#define PD_GET_ADDRESS  0x0600

#define PD_RCV_MODE     0x1400  /* AU Monitor */
#define PM_DEFAULT      3  /* Unicast + Broadcast (Nevil, 27 Dec 96) */
#define PM_PROMISCUOUS  6

#define PD_GET_PARAMS   0x0A00  /* Nevil, 15 Jul 96 */
struct pd_param {
   unsigned char major_rev;
   unsigned char minor_rev;
   unsigned char length;
   unsigned char addr_len;
   unsigned short mtu;
   unsigned short multicast_aval;
   unsigned short rcv_bufs;
   unsigned shortxmt_bufs;
   unsigned short int_num;
   };

#define  CARRY          1            /* carry bit in flags register */

Bit16 _pktipofs = 0;                    /* offset from header to start of pkt */
Bit16 pkt_interrupt;
Bit16 pkt_ip_type = 0x0008;             /* these are intelled values */
Bit16 pkt_arp_type = 0x0608;

struct ether outbuf_mem;  /* Allocated in near memory */
struct ether *outbuf = &outbuf_mem;
struct pbuf pktbuf_mem[MAXBUFS];
typedef byte (*pbuf_ptr)[sizeof(struct pbuf)];  /* Packet buffers */
pbuf_ptr pktbuf = (pbuf_ptr)&pktbuf_mem;

Bit16 pkt_arp_handle;
Bit8  eth_addr[6];

long rx_packets = 0, rx_bytes = 0,  /* AU monitor */
   tx_packets = 0, tx_bytes = 0,
   found_packets = 0,
   badpackets = 0, nobufpackets = 0,
   send_fails = 0,  // Nevil, 10 Jan 97
   low_send_fails = 0;  // Joel, 22 Jan 97

Bit32 far *interrupts = 0L;
char *pkt_line = "PKT DRVR";

#define IP_DF       0x0040      // Don't fragment bit set for FRAG Flags

// forwards and externs
byte *fragment( in_Header *ip );

int pkt_init(void);
struct en_info en_table[MAX_PD] =  /* Default: one pkt driver, IP only */
   {{EN_IP,0,0}, {0,0,0}, {0,0,0}, {0,0,0}};

struct bufq freeq;
struct bufq readyq;

/* Add a pbuf to the tail of a queue */
void enqueue(struct bufq *bq, struct pbuf *pb)
{
   asm pushf;  /* Save flags (especially int flag) */
   asm cli;    /* Disable interrupts */
      if (bq->len == 0) bq->head = pb;
      else bq->tail->next = pb;
      bq->tail = pb;
      ++bq->len;
   asm popf;   /* Restore flags */
   }

/* Put a pbuf into a queue, maintained in ascending seq order */
void seq_enqueue(struct bufq *bq, struct pbuf *np)
{
   struct pbuf *p, *prev_p;
   asm pushf;  /* Save flags (especially int flag) */
   asm cli;    /* Disable interrupts */
      if (bq->len == 0) {  /* Empty queue */
         bq->head = bq->tail = np;  np->next = NULL;
         }
      else {
         p = bq->head;
         if (np->seq < p->seq) {  /* Add to head */
            bq->head = np;  np->next = p;
            }
         else for (;;) {
            prev_p = p;  p = p->next;
            if (p == NULL) {  /* Add to tail */
               prev_p->next = np;  np->next = NULL;
               bq->tail = np;
               break;
               }
            if (np->seq < p->seq) {  /* Add into queue */
               prev_p->next = np;  np->next = p;
               break;
               }
            }
         }
      ++bq->len;
   asm popf;   /* Restore flags */
   }

/* Take the head pbuf from a queue */
struct pbuf *dequeue(struct bufq *bq)
{
   struct pbuf *rb;
   asm pushf;  /* Save flags (especially int flag) */
   asm cli;    /* Disable interrupts */
      if (bq->len == 0) rb = NULL;
      else {
         rb = bq->head;
         bq->head = rb->next;
         --bq->len;
         }
   asm popf;   /* Restore flags */
   return rb;
   }

void pd_error(unsigned int err_nbr, int sw_int, char *msg)
{
   outs("PD ERROR 0x");  outhex(err_nbr >> 8);
   outs(", interrupt 0x");  outhex(sw_int);
   outs(": ");  outs(msg);
   }

int pkt_rcv_mode(int pkt_int, int handle, int mode)
{
   struct REGPACK regs;
   regs.r_ax = PD_RCV_MODE;
   regs.r_bx = handle;
   regs.r_cx = mode;
   intr(pkt_int, &regs);
   if (regs.r_flags & CARRY) {
      pd_error(regs.r_dx, pkt_int, "setting receive mode\n\r");
      return 1;
      }
   return 0;
   }

typedef void (*fn_type)();

int check_pkt_driver(int *pi,  /* Returns 0 if OK, 1 otherwise */
   fn_type pkt_entry_real_ptr)
{
   struct REGPACK regs;
   char far *temp;
   int pd_ok, pint, pd_type, class;

#if PKT_DEBUG
   printf("   check_pkt_driver(%d)\n", *pi);
#endif
   pd_ok = 0;
   if (*pi == 0) {
      for (pint = INT_FIRST; pint <= INT_LAST; ++pint) {
         temp = (char far *)(interrupts[pint]);
         if (!memcmp(&(temp[3]), pkt_line, strlen(pkt_line))) {
            *pi = pint;
            pd_ok = 1;
            break;
            }
         }
      }
   else {
      temp = (char far *)(interrupts[pint = *pi]);
      if (!memcmp(&(temp[3]), pkt_line, strlen(pkt_line)))
         pd_ok = 1;
      }
   if (!pd_ok) {
      pd_error(0, *pi, "NO PACKET DRIVER FOUND\r\n");
      return 1;
      }

   /* Find out about the driver */
   regs.r_ax = PD_DRIVER_INFO;
   intr(pint, &regs);
   /* Handle old versions, assume a class and just keep trying */
   if (regs.r_flags & CARRY) {
      for (class = 0; class < 2; ++class) {
         _pktdevclass = (class) ? PD_SLIP : PD_ETHER;
         for (pd_type = 1; pd_type < 128; ++pd_type) {
            regs.r_ax = PD_ACCESS | _pktdevclass;  /* ETH, SLIP */
            regs.r_bx = pd_type;                /* type */
            regs.r_dx = 0;                      /* if number */
            regs.r_cx = (_pktdevclass == PD_SLIP) ? 0 : sizeof(pkt_ip_type);
            regs.r_ds = FP_SEG(&pkt_ip_type);
            regs.r_si = FP_OFF(&pkt_ip_type);
            regs.r_es = FP_SEG(pkt_entry_real_ptr);
            regs.r_di = FP_OFF(pkt_entry_real_ptr);
            intr(pint, &regs);
            if (!(regs.r_flags & CARRY)) break;
            }
         if (pd_type == 128 ) {
            pd_error(0, pint, "couldn't initialise driver\n\r");
            return 1;
            }
         /* we have found a working type, so kill it */
         regs.r_bx = regs.r_ax; /* handle */
         regs.r_ax = PD_RELEASE;
         intr(pint, &regs);
         }
      }
   else {
      pd_type = regs.r_dx;
      switch (_pktdevclass = (regs.r_cx >> 8)) {
      case PD_ETHER:
         _pktipofs = 14;
      case PD_SLIP:
         break;
      default:
         pd_error(0, pint, "only Ethernet or SLIP drivers allowed\n\r");
         return 1;
         }
      }
#if PKT_DEBUG
   printf("      pint=0x%02X, _pktipofs=%d\n", pint,_pktipofs);
#endif
   return 0;
   }

int read_en_address(int if_nbr)
{
   int pkt_int;
   struct REGPACK regs;

   pkt_int = en_table[if_nbr].int_nbr;
   regs.r_ax = PD_GET_ADDRESS;
   regs.r_bx = en_table[if_nbr].handle;
   regs.r_es = FP_SEG(eth_addr);
   regs.r_di = FP_OFF(eth_addr);
   regs.r_cx = sizeof(eth_addr);
   intr(pkt_int, &regs);
   if (regs.r_flags & CARRY) {
      pd_error(regs.r_dx, pkt_int, "reading Ethernet address\n\r");
      return 1;
      }
   return 0;
   }

int en_init(int if_nbr)  /* Returns 0 if OK, 1 otherwise */
{
   int pkt_int, if_type;
   struct REGPACK regs, regs2;
   struct pd_param far *pd_pp;  /* Nevil, 15 Jul 95 */
   int zero = 0;  /* AU */
   fn_type pkt_ent[] = {
      pkt_ent_0, pkt_ent_1, pkt_ent_2, pkt_ent_3 };

   if (check_pkt_driver(&en_table[if_nbr].int_nbr, pkt_ent[if_nbr]))
      return 1;  /* Fail */

   pkt_int = en_table[if_nbr].int_nbr;  /* May be set by check_pkt (above) */
   if_type = en_table[if_nbr].type;
   if (if_type & EN_IP)  /* IP on this interface */
      pkt_interrupt = pkt_int;
#if PKT_DEBUG
   printf("   en_init(%d), pkt_int=0x%02X, type=%d, pkt_interrupt=0x%02X\n",
      if_nbr,pkt_int,if_type, pkt_interrupt);
#endif

   if (if_type & EN_HP) { /* Could use high-performance driver */
      /* Get driver parameters */
      regs.r_ax = PD_GET_PARAMS;
      intr(pkt_int, &regs);
      if ((regs.r_flags & CARRY) == 0) {  /* Success */
         pd_pp = MK_FP(regs.r_es,regs.r_di);
         if (pd_pp->minor_rev < 10) if_type ^= EN_HP;
         }
      else if_type ^= EN_HP;
      }

   /* Tell the packet driver our callback address and desired ethertype */
   regs.r_ax = PD_ACCESS | _pktdevclass;
   regs.r_bx = 0xffff;   /* Any type */
   regs.r_dx = 0;  /* if_number */
   regs.r_es = FP_SEG(pkt_ent[if_nbr]);
   regs.r_di = FP_OFF(pkt_ent[if_nbr]);

   if (if_type & EN_MT) {  /* Metering required */
      regs.r_cx = 0;  /* typelen */
      regs.r_ds = FP_SEG(&zero);
      regs.r_si = FP_OFF(&zero);
      intr(pkt_int, &regs);
      if (regs.r_flags & CARRY) {
         pd_error(regs.r_dx, pkt_int, "accessing packet driver\n\r");
         return 1;
         }

      en_table[if_nbr].handle = regs.r_ax;
      if (if_type & EN_IP) {  /* IP on this interface */
         au_monitor = 1;
         read_en_address(if_nbr);
         }
      pkt_rcv_mode(pkt_int, en_table[if_nbr].handle, PM_PROMISCUOUS);
      return 0;  /* OK */
      }

   /* No metering: assume IP is required */
   regs.r_cx = (_pktdevclass == PD_SLIP) ? 0 : sizeof(pkt_ip_type);
   regs.r_ds = FP_SEG(&pkt_ip_type);
   regs.r_si = FP_OFF(&pkt_ip_type);
   memcpy(&regs2, &regs, sizeof(regs));
   regs2.r_ds = FP_SEG(&pkt_arp_type);
   regs2.r_si = FP_OFF(&pkt_arp_type);
   intr(pkt_interrupt, &regs);
   if (regs.r_flags & CARRY) {
      pd_error(regs.r_dx, pkt_int, "accessing packet driver for IP\n\r");
      return 1;
      }
   en_table[if_nbr].handle = regs.r_ax;
   read_en_address(if_nbr);
#if PKT_DEBUG
   printf("      IP handle  = 0x%04X\n", regs.r_ax);
#endif
   if (_pktdevclass != PD_SLIP) {  /* Get ARP handle */
      intr(pkt_interrupt, &regs2);
      if (regs2.r_flags & CARRY) {
         regs.r_ax = PD_RELEASE;  /* Release IP handle */
         regs.r_bx = en_table[if_nbr].handle;
         intr(pkt_int, &regs);
         pd_error(regs.r_dx, pkt_int, "accessing driver for ARP\n\r" );
         return 1;
         }
      pkt_arp_handle = regs2.r_ax;
      }
#if PKT_DEBUG
   printf("      ARP handle = 0x%04X\n", pkt_arp_handle);
#endif
   return 0;  /* OK */
   }

int pkt_init(void)
{
   int j;
   pkt_buf_wipe();  /* Initialise ethernet buffer pool */
   for (j = 0; j != MAX_PD; ++j) {
      if (en_table[j].type) {
         if (en_init(j)) return 1;
         }
      }
   /* hwt_init();  Now initiliased in timestmp.cpp.  Nevil, 8 Jun 98 */
   return 0;  /* OK */
   }

void en_release(int if_nbr)
{
   int pkt_int = en_table[if_nbr].int_nbr;
   int handle  = en_table[if_nbr].handle;
   int if_type = en_table[if_nbr].type;
   struct REGPACK regs;

   if (if_type & EN_MT) {  /* We were metering */
      pkt_rcv_mode(pkt_int, handle, PM_DEFAULT);
      }
   else if (if_type & EN_IP) {  /* IP without metering */
      if (_pktdevclass != PD_SLIP) {
         regs.r_ax = PD_RELEASE;
         regs.r_bx = pkt_arp_handle;
         intr(pkt_int, &regs);
         if (regs.r_flags & CARRY)
         pd_error(regs.r_dx, pkt_int, "releasing driver for ARP\n\r");
         }
      }
   regs.r_ax = PD_RELEASE;
   regs.r_bx = handle;
   intr(pkt_int, &regs);
   if (regs.r_flags & CARRY)
      pd_error(regs.r_dx, pkt_int, "releasing driver\n\r");
   }

void pkt_release()
{
   int j;
   for (j = 0; j != MAX_PD; ++j) {
      if (en_table[j].int_nbr)
         en_release(j);
      }
   hwt_fin();  // Finished with clock routines
   }

unsigned char *get_pktbuf(Bit16 length_wanted)
{
   struct pbuf *bp;

   // Make sure the packet driver isn't asking for more than one buffer
   if (length_wanted > ETH_MTU || length_wanted == 0) {
      ++badpackets;
#if GETBUF_TALKS
      printf("get_pktbuf() failed; pkt driver wanted more than %d bytes\n",
         BUFSIZE);
#endif
      return (Bit8 *)NULL;
      }

   if ((bp = dequeue(&freeq)) != NULL) {
      bp->state = 1;  /* Waiting for 2nd interrupt */
      ++rx_packets;  rx_bytes += length_wanted;
      return (Bit8 *)&bp->data;
      }

#if GETBUF_TALKS
      printf("get_pktbuf() failed; no buffers free\n");
#endif
   rx_bytes += length_wanted;  /* Don't loose the counts */
   return (Bit8 *)NULL;
   }

int pkt_send(char *buffer, int length)
{
   struct REGPACK regs;
   int retries;
   unsigned char au_save;

   retries = 500;  // Joel, Feb 97
   // Using a big number of retries means we busy-wait here,
   // hoping the ethernet card can send the packet real soon now
   while (retries--) {
      regs.r_ax = PD_SEND;
      regs.r_ds = FP_SEG(buffer);
      regs.r_si = FP_OFF(buffer);
      regs.r_cx = length;
      intr(pkt_interrupt, &regs);
      if (regs.r_flags & CARRY) {  /* Send failed */
         ++low_send_fails;  // Joel, 22 Jan 97
         continue;
         }

      ++tx_packets;  // Update counters
      tx_bytes += length;
      if (au_monitor)  /* Monitor transmitted packets as well as received */
         au_send_mon(length, (unsigned char far *)buffer);
      return 0 ;
      }
   ++send_fails;  // Nevil, 10 Jan 97
   return  1;  /* Failed to send */
   }

void pkt_buf_wipe()  /* Zero out all buffers */
{
   int j;
   struct pbuf *bp;
   freeq.len = readyq.len = 0;
   for (j = 0; j != MAXBUFS; ++j) {  // Link buffers into free queue
      bp = (struct pbuf *)&pktbuf[j];
      bp->unused = j;  // 'unused' byte = buffer number
      bp->state = 0;
      enqueue(&freeq, bp);
      }
   }

void pkt_buf_release(char *ptr)  /* Return one IP buffer to the pool */
{
   struct pbuf *bp;
   bp = (struct pbuf *)(ptr - _pktipofs - BUF_HDR_LEN);
   bp->state = 0;
#if IP_BUF_TRACE
   printf(" {-%u}", bp->unused);
#endif
   enqueue(&freeq, bp);
   }


void *pkt_received(struct bufq *bq)
{
   struct pbuf *bp;
   bp = dequeue(bq);
   if (bp == NULL) return NULL;  /* No packets ready */

   ++found_packets;
   bp->state = 3;  /* In use by application */
#if IP_BUF_TRACE
   printf(" {+%u}", bp->unused);
#endif
   return &bp->data;
   }

void *_pkt_eth_init()
{
   if (pkt_init()) {
      outs("Program halted\n");
      exit(1);
      }
   return eth_addr;
   }
