#include "../intel/timestmp.h"



//
//  Timer routines - use set_*timeout to receive a clock value which
//                   can later be tested by calling chk_timeout with
//                   that clock value to determine if a timeout has
//                   occurred.  chk_timeout returns 0 if NOT timed out.
//
//
//  Usage :
//           long set_timeout( int seconds )
//           long set_ttimeout( int pc_ticks )
//           int chk_timeout( long value )
//




// since a 32-bit number can only count 7.5 years with a 18.2 Hz timebase,
// and since UNIX uses absolute times from 1980, we must subtract some
// more recent baseline time from each timestamp, to avoid having them
// all max out at 2^32-1
//
// not necessary on a Pentium because it counts from reset, not realtime
//
static PentiumClock init_timeout( void );
static PentiumClock startup = init_timeout();

static PentiumClock init_timeout( void )
{
    startup.set();
    return startup;
}



extern "C"
unsigned long set_timeout( int seconds )
{
    PentiumClock temp, now;

    now.set();
    now -= startup;
    temp.put_seconds( seconds );
    temp += now;
    return temp.get_ticks();
}



extern "C"
unsigned long set_ttimeout( int pc_ticks )
{
    PentiumClock temp, now;

    now.set();
    now -= startup;
    temp.put_ticks( pc_ticks );
    temp += now;
    return temp.get_ticks();
}



extern "C"
int chk_timeout( unsigned long value )
{
    PentiumClock temp, now;

    now.set();
    now -= startup;
    temp.put_ticks( value );
    return now >= temp;
}
